#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from modusa.generators.base import ModusaGenerator
from modusa.models.t_ax import TAx
import numpy as np

class TAxGen(ModusaGenerator):
	"""
	Provides user friendly APIs to generate time
	axis for signals (instances of `TAx`).
	"""
	
	#--------Meta Information----------
	_name = "TimeAxisGenerator"
	_description = "APIs to generate time axis for signals."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-26"
	#----------------------------------
		
	@classmethod
	def linear(cls, n_points: int, sr: int | float = 1.0, t0: int | float = 0.0, label: str = "Time (sec)") -> TAx:
		"""
		Create a linearly spaced time axis.
		
		.. code-block:: python

			import modusa as ms
			t = ms.tax.linear(n_points=100, sr=2, start=10, label="Time (sec)")
			print(t)
			t.print_info()
		
		Parameters
		----------
		n_points: int
			- Number of data points for the time axis.
		sr: int | float
			- Sampling rate for the time axis.
		start: int | float
			- Start value.
		label: str
			- Label for the time axis.
		
		Returns
		-------
		TAx
			An instance of TAx.
		"""
		
		assert isinstance(n_points, int)
		assert isinstance(sr, (int, float))
		assert isinstance(t0, (int, float))
		assert isinstance(label, str)
		
		sr = float(sr)
		t0 = float(t0)
		
		time_ax = TAx(n_points=n_points, sr=sr, t0=t0, label=label)
		
		return time_ax
	


#!/usr/bin/env python3

from abc import ABC, abstractmethod

class ModusaTool(ABC):
	"""
	Base class for all tool: youtube downloader, audio converter, filter.
	
	>>> modusa-dev create io

	.. code-block:: python
		
		# General template of a subclass of ModusaTool
		from modusa.tools.base import ModusaTool

		class MyCustomIOClass(ModusaIO):
			#--------Meta Information----------
			_name = "My Custom Tool"
			_description = "My custom class for Tool."
			_author_name = "Ankit Anand"
			_author_email = "ankit0.anand0@gmail.com"
			_created_at = "2025-07-06"
			#----------------------------------
			
			@staticmethod
			def do_something():
				pass
		
	
	Note
	----
	- This class is intended to be subclassed by any tool built for the modusa framework.
	- In order to create a tool, you can use modusa-dev CLI to generate a template.
	- It is recommended to treat subclasses of ModusaTool as namespaces and define @staticmethods with control parameters, rather than using instance-level __init__ methods.
	"""
	
	#--------Meta Information----------
	_name: str = "Modusa Tool"
	_description: str = "Base class for any tool in the Modusa framework."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-11"
	#----------------------------------


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .s1d import S1D
from .t_ax import TAx
from .data import Data
from modusa.tools.math_ops import MathOps
from typing import Self, Any, Callable
from types import NoneType
import numpy as np
import matplotlib.pyplot as plt

class TDS(S1D):
	"""
	Space to represent time domain signals.
	
	Note
	----
	- Use :class:`~modusa.generators.tds.TDSGen` to instantiate this class.

	Parameters
	----------
	y: Data
		- Data object holding the main array.
	t: TAx
		- Time axis for the signal.
	title: str
		- Title for the signal.
		- Default: None => ''
		- e.g. "MySignal"
		- This is used as the title while plotting.
	"""
	
	#--------Meta Information----------
	_name = "Time Domain Signal"
	_nickname = "signal" # This is to be used in repr/str methods
	_description = "Space to represent uniform time domain signal."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-20"
	#----------------------------------
	

	def __init__(self, y, t, title = None):
		
		if not (isinstance(y, Data) and isinstance(t, TAx)):
			raise TypeError(f"`y` must be `Data` instance and `t` must be `TAx` object, got {type(y)} and {type(x)}")
			
		assert y.ndim == 1
		
		super().__init__(y=y, x=t, title=title) # Instantiating `Signal1D` class
	
	#---------------------------------
	# Properties (Hidden)
	#---------------------------------
	
	@property
	def y(self) -> Data:
		return self._y
	
	@property
	def t(self) -> TAx:
		return self.x
	
	@property
	def title(self) -> str:
		return self._title
	
	@property
	def shape(self) -> tuple:
		return self.y.shape
	
	@property
	def ndim(self) -> tuple:
		return self.y.ndim # Should be 1
	
	@property
	def size(self) -> int:
		return self.y.size
	
	#==================================
	
	#-------------------------------
	# NumPy Protocol
	#-------------------------------
	
	def __array__(self, dtype=None) -> np.ndarray:
		return np.asarray(self.y.values, dtype=dtype)
	
	def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
		"""
		Supports NumPy universal functions on the Signal1D object.
		"""
		from .data import Data  # Ensure this is the same Data class you're using
		from modusa.utils import np_func_cat as nfc
		
		raw_inputs = [
			np.asarray(obj.y) if isinstance(obj, type(self)) else obj
			for obj in inputs
		]
		
		result = getattr(ufunc, method)(*raw_inputs, **kwargs)
		
		y = Data(values=result, label=None)  # label=None or you could copy from self.y.label
		t = self.t.copy()
		
		if y.shape != t.shape:
			raise ValueError(f"`{ufunc.__name__}` caused shape mismatch between data and axis, please create a github issue")
			
		return self.__class__(y=y, t=t, title=self.title)
	
	def __array_function__(self, func, types, args, kwargs):
		"""
		Additional numpy function support.
		"""
		from modusa.utils import np_func_cat as nfc
		
		if not all(issubclass(t, type(self)) for t in types):
			return NotImplemented
		
		# Not supporting concatenate like operations as axis any random axis can't be concatenated
		if func in nfc.CONCAT_FUNCS:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
			
		# Single signal input expected
		signal = args[0]
		result: Data = func(signal.y, **kwargs)
		axis = kwargs.get("axis", None)
		keepdims = kwargs.get("keepdims", None)
		
		if func in nfc.REDUCTION_FUNCS:
			if keepdims is None or keepdims is True: # Default state is True => return 1D signal by wrapping the scalar
				from .t_ax import TAx
				dummy_t = TAx(n_points=1, sr=signal.t.sr, t0=0, label=None)
				return self.__class__(y=result, t=dummy_t, title=signal.title)
			elif keepdims is False: # Return Data
				from .data import Data
				return Data(values=result, label=None)
		
		elif func in nfc.X_NEEDS_ADJUSTMENT_FUNCS:
			# You must define logic for adjusting x
			raise NotImplementedError(f"{func.__name__} requires x-axis adjustment logic.")
			
		else:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
		
	#================================
	
	#-------------------------------
	# Indexing
	#-------------------------------
			
	def __getitem__(self, key):
		"""
		Return a sliced or indexed view of the data.
		
		Parameters
		----------
		key : array-like
			- Index to apply to the values.
		
		Returns
		-------
		TDS
			A new TDS object with sliced values and same meta data.
		"""
		if not isinstance(key, (int, slice)):
			raise TypeError(f"Invalid key type: {type(key)}")
		
		sliced_y = self.y[key]
		sliced_t = self.t[key]
		
		if sliced_y.ndim == 0:
			sliced_y = Data(values=sliced_y.values, label=sliced_y.label, ndim=1)
			
		return self.__class__(y=sliced_y, t=sliced_t, title=self.title)
	
	def __setitem__(self, key, value):
		"""
		Set values at the specified index.
	
		Parameters
		----------
		key : int | slice | array-like | boolean array | S1D
			Index to apply to the values.
		value : int | float | array-like
			Value(s) to set.
		"""
		
		self.y[key] = value  # In-place assignment
		
	#===================================
	
	
	#-----------------------------------
	# Utility Methods
	#-----------------------------------
	
	def unpack(self):
		"""
		Unpacks the object into easy to work
		with data structures.

		Returns
		-------
		(np.ndarray, float, float)
			- y: Signal data array.
			- sr: Sampling rate of the signal.
			- t0: Starting timestamp.
		"""
		
		arr = self.y.values
		sr = self.t.sr
		t0 = self.t.t0
		
		return (arr, sr, t0)
		
	def copy(self) -> Self:
		"""
		Returns a new copy of the signal.
		
		Returns
		-------
		Self
			A new copy of the object.
		"""
		copied_y = self.y.copy()
		copied_x = self.x.copy()
		title = self.title # Immutable, hence no need to copy
		
		return self.__class__(y=copied_y, x=copied_x, title=title)
		
	
	def set_meta_info(self, title = None, y_label = None, t_label = None) -> None:
		"""
		Set meta info about the signal.

		Parameters
		----------
		title: str
			- Title for the signal
			- e.g. "Speedometer"
		y_label: str
			- Label for the y-axis.
			- e.g. "Speeed (m/s)"
		t_label: str
			- Label for the time-axis.
			- e.g. "Distance (m)"
		"""
		
		new_title = str(title) if title is not None else self.title
		new_y_label = str(y_label) if y_label is not None else self.y.label
		new_t_label = str(t_label) if t_label is not None else self.t.label
		
		# We create a new copy of the data and axis
		new_y = self.y.copy().set_meta_info(y_label)
		new_t = self.t.copy().set_meta_info(t_label)
		
		return self.__class__(y=new_y, t=new_t, title=title)
	
	
	def is_same_as(self, other: Self) -> bool:
		"""
		Check if two `TDS` instances are equal.
		"""
		
		if not isinstance(other, type(self)):
			return False
		
		if not self.y.is_same_as(other.y):
			return False
		
		if not self.t.is_same_as(other.t):
			return False
		
		return True
	
	def has_same_axis_as(self, other) -> bool:
		"""
		Check if two 'TDS' instances have same
		axis. Many operations need to satify this.
		"""
		return self.t.is_same_as(other.t)
	
	
	def mask(self, condition, set_to=None) -> Self:
		"""
		Mask the signal based on condition and
		the values can be set.
		
		Parameters
		----------
		condition: Callable
			- Condition function to apply on values of the signal.
			- E.g. lambda x: x > 10
		set_to: Number
			- Number to replace the masked position values.

		Returns
		-------
		TDS
			Masked Signal
		"""
		
		mask = condition(self)
		new_val = set_to
		
		if set_to is None: # Return the mask as the same signal but with booleans
			return mask
	
		else:
			# We apply the mask and update the signal data
			new_data = self.y.mask(condition=condition, set_to=new_val)
			
			# Since we're just updating the data, there is no change in the axis
			return self.__class__(y=new_data, t=self.t.copy(), title=self.title)
	#===================================
	
	#-------------------------------
	# Tools
	#-------------------------------
		
	@validate_args_type()
	def translate_t(self, n_samples: int):
		"""
		Translate the signal along time axis.
		
		Note
		----
		- Negative indexing is allowed but just note that you might end up getting time < 0


		.. code-block:: python
			
			import modusa as ms
			s1 = ms.tds([1, 2, 4, 4, 5, 3, 2, 1])
			ms.plot(s1, s1.translate_t(-1), s1.translate_t(3))
		
		Parameters
		----------
		n_samples: int
			By how many sample you would like to translate the signal.
		
		Returns
		-------
		TDS
			Translated signal.
		"""
		
		translated_t = self.t.translate(n_samples=n_samples)
		
		return self.__class__(y=self.y.copy(), t=translated_t, title=self.title)
	
	def pad(self, left=None, right=None) -> Self:
		"""
		Pad the signal with array like object from the
		left or right.

		Parameters
		----------
		left: arraylike
			- What to pad to the left of the signal.
			- E.g. 1 or [1, 0, 1], np.array([1, 2, 3])
		right: arraylike
			- What to pad to the right of the signal.
			- E.g. 1 or [1, 0, 1], np.array([1, 2, 3])

		Returns
		-------
		TDS
			Padded signal.
		"""
		
		if right is None and left is None: # No padding applied
			return self
		
		# Pad the data
		y_padded = self.y.pad(left=left, right=right)
			
		# Find the new t0
		if left is not None:
			if np.ndim(left) == 0: left = np.asarray([left])
			else: left = np.asarray(left)
			new_t0 = self.t.t0 - (left.shape[0] / self.t.sr)
		else:
			new_t0 = self.t.t0
		
		t_padded = self.t.__class__(n_points=y_padded.shape[0], sr=self.t.sr, t0=new_t0, label=self.t.label)
		
		return self.__class__(y=y_padded, t=t_padded, title=self.title)
		
		
	
	def crop(self, t_min = None, t_max = None, like = None) -> Self:
		"""
		Crop the signal to a time range [t_min, t_max].

		.. code-block:: python

			import modusa as ms
			s1 = ms.tds.random(1000, sr=10)
			ms.plot(s1, s1.crop(5, 40), s1.crop(20), s1.crop(60, 80))

	
		Parameters
		----------
		t_min : float or None
			Inclusive lower time bound in second (other units). If None, no lower bound.
		t_max : float or None
			Exclusive upper time bound in second (other units). If None, no upper bound.
		like: TDS
			- A `TDS` object whose start and end time will be used.
			- If you have a window signal, you can crop the signal to get the correct portion.
		
		Returns
		-------
		TDS
			Cropped signal.
		"""
		
		if like is not None:
			ref_signal = like
			assert self.t.sr == ref_signal.t.sr
			# Set t_min and t_max as per the signal
			t_min = ref_signal.t.t0
			t_max = ref_signal.t.end_time
		
		# We first will find out the time in samples
		if t_min is not None:
			t_min_sample = self.t.index_of(t_min)
		else:
			t_min_sample = 0
			
		if t_max is not None:
			t_max_sample = self.t.index_of(t_max)
		else:
			t_max_sample = -1
		
		return self[t_min_sample: t_max_sample+1]
	
	#===================================
	
	#-----------------------------------
	# Information
	#-----------------------------------
	
	def print_info(self) -> None:
		"""Prints info about the audio."""
		print("-" * 50)
		print(f"{'Title'}: {self.title}")
		print("-" * 50)
		print(f"{'Type':<20}: {self.__class__.__name__}")
		print(f"{'Shape':<20}: {self.shape}")
		print(f"{'Duration':<20}: {self.t.duration:.2f} sec")
		print(f"{'Sampling Rate':<20}: {self.t.sr} Hz")
		print(f"{'Sampling Period':<20}: {(1 / self.t.sr * 1000):.2f} ms")
		
		# Inheritance chain
		cls_chain = " → ".join(cls.__name__ for cls in reversed(self.__class__.__mro__[:-1]))
		print(f"{'Inheritance':<20}: {cls_chain}")
		print("=" * 50)
	
	#======================================


#!/usr/bin/env python3


import soundfile as sf
from scipy.signal import resample
from pathlib import Path
import tempfile
from scipy.signal import resample
from .youtube_downloader import download
from .audio_converter import convert


def load(path, sr=None, clip=None):
	"""
	Loads audio file from various sources.

	.. code-block:: python
		
		import modusa as ms
		audio_fp = ms.load(
			"https://www.youtube.com/watch?v=lIpw9-Y_N0g",
			sr = None, clip=(5, 10))

	Parameters
	----------
	path: str
		- Path to the audio
		- Youtube URL
	sr: int | None
		- Sampling rate to load the audio in.
	clip: number | tuple[number, number] | None
		- Which segment of the audio you want.
		- Eg., 10 => First 10 sec, (5, 10) => 5 to 10 second
		- Default: None => Entire audio.

	Return
	------
	np.ndarray
		- Audio signal.
	int
		- Sampling rate of the loaded audio signal.
	title
		- Title of the loaded audio.
		- Filename without extension or YouTube title.
	"""
	
	# Check if the path is YouTube
	if ".youtube." in str(path):
		# Download the audio in temp directory using tempfile module
		with tempfile.TemporaryDirectory() as tmpdir:
			# Download
			audio_fp: Path = download(url=path, content_type="audio", output_dir=Path(tmpdir))
			
			# Convert the audio to ".wav" form for loading
			wav_audio_fp: Path = convert(inp_audio_fp=audio_fp, output_audio_fp=audio_fp.with_suffix(".wav"))
			
			# Load the audio in memory
			audio_data, audio_sr = sf.read(wav_audio_fp)
			title = audio_fp.stem
			
			# Convert to mono if it's multi-channel
			if audio_data.ndim > 1:
				audio_data = audio_data.mean(axis=1)
				
			# Resample if needed
			if sr is not None:
				if audio_sr != sr:
					n_samples = int(len(audio_data) * sr / audio_sr)
					audio_data = resample(audio_data, n_samples)
					audio_sr = sr
		
	else:
		# Check if the file exists
		fp = Path(path)
		
		if not fp.exists():
			raise FileNotFoundError(f"{path} does not exist.")
		
		# Load the audio in memory
		audio_data, audio_sr = sf.read(fp)
		title = fp.stem
		
		# Convert to mono if it's multi-channel
		if audio_data.ndim > 1:
			audio_data = audio_data.mean(axis=1)
			
		# Resample if needed
		if sr is not None:
			if audio_sr != sr:
				n_samples = int(len(audio_data) * sr / audio_sr)
				audio_data = resample(audio_data, n_samples)
				audio_sr = sr
				
	# Clip the audio signal as per needed
	if clip is not None:
		# Map clip input to the right format
		if isinstance(clip, int or float):
			clip = (0, clip)
		elif isinstance(clip, tuple) and len(clip) > 1:
			clip = (clip[0], clip[1])
		else:
			raise ValueError(f"Invalid clip type or length: {type(clip)}, len={len(clip)}")
		
		audio_data = audio_data[int(clip[0]*audio_sr):int(clip[1]*audio_sr)]
	
	return audio_data, audio_sr, title


#!/usr/bin/env python3

import logging

class Config:
	LOG_LEVEL = logging.WARNING
	SR = 44100  # Default sampling rate
	TIME_UNIT = "sec"
	
	
	def __str__(self):
		return self.__dict__
	
	def __repr__(self):
		return self.__dict__

# Create a singleton instance
config = Config()


#!/usr/bin/env python3

from .base import ModusaGenerator

from .s_ax import SAxGen
from .t_ax import TAxGen

from .s1d import S1DGen
from .tds import TDSGen
from .audio import AudioGen

from .s2d import S2DGen
from .ftds import FTDSGen


#!/usr/bin/env python3

#---------------------------------
# Author: Ankit Anand
# Date: 12/08/25
# Email: ankit0.anand0@gmail.com
#---------------------------------

def load_ann(path, clip=None):
	"""
	Load annotation from audatity label
	text file and also ctm file.

	Parameters
	----------
	path: str
		- label text/ctm file path.
	clip: tuple[number, number] | number | None
		- Incase you clipped the audio signal, this parameter will help clip the annotation.
		- If you clip the audio, say from (10, 20), set the clip to (10, 20).
		- Default: None

	Returns
	-------
	list[tuple, ...]
		- annotation data structure
		- [(start, end, label), ...]
	"""
	from pathlib import Path
	
	if not isinstance(path, (str, Path)):
		raise ValueError(f"`path` must be one of (str, Path), got {type(path)}")
	
	# Convert to Path object
	path = Path(path)
	
	# Check if the path exists
	if not path.exists():
		raise FileExistsError(f"{path} does not exist")
		
	ann = [] # This will store the annotation
	
	# Clipping the annotation to match with the clipped audio
	if clip is not None:
		# Map clip input to the right format
		if isinstance(clip, int or float):
			clip = (0, clip)
		elif isinstance(clip, tuple) and len(clip) > 1:
			clip = (clip[0], clip[1])
		else:
			raise ValueError(f"Invalid clip type or length: {type(clip)}, len={len(clip)}")
	
	if path.suffix == ".txt":
		with open(str(path), "r") as f:
			lines = [line.rstrip("\n") for line in f]
			for line in lines:
				start, end, label = line.split("\t")
				start, end = float(start), float(end)
				
				# Incase user has clipped the audio signal, we adjust the annotation
				# to match the clipped audio
				if clip is not None:
					offset = clip[0]
					# Clamp annotation to clip boundaries
					new_start = max(start, clip[0]) - offset
					new_end   = min(end, clip[1]) - offset
					
					# only keep if there's still overlap
					if new_start < new_end:
						ann.append((new_start, new_end, label))
				else:
					ann.append((start, end, label))
					
	elif path.suffix == ".ctm":
		with open(str(path), "r") as f:
			content = f.read().split("\n")
			
		for c in content:
			c = c.strip()
			if c == "": # Handle empty line usually at the end of the ctm file
				continue
			elif len(c.split(" ")) != 5:
				warnings.warn(f" '{c}' is not a standard ctm line.")
			if len(c.split(" ")) == 5:
				_, _, start, dur, label = c.split(" ")
				start, dur = float(start), float(dur)
				end = start + dur
				
				# Incase user has clipped the audio signal, we adjust the annotation
				# to match the clipped audio
				if clip is not None:
					offset = clip[0]
					# Clamp annotation to clip boundaries
					new_start = max(start, clip[0]) - offset
					new_end   = min(end, clip[1]) - offset
					
					# only keep if there's still overlap
					if new_start < new_end:
						ann.append((new_start, new_end, label))
				else:
					ann.append((start, end, label))
				
	else:
		raise Exception(f"Unsupported file type {path.suffix}")
	return ann


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from .base import ModusaGenerator
from modusa.models.s1d import S1D
from modusa.models.s_ax import SAx
from modusa.models.data import Data
import numpy as np

class S1DGen(ModusaGenerator):
	"""
	Provides user friendly APIs to generate instances of different `S1D`
	instances.
	"""
	
	#--------Meta Information----------
	_name = ""
	_description = ""
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-27"
	#----------------------------------
	
	@staticmethod
	def from_array(
		y,
		x = None,
		y_label: str = "Y",
		x_label: str = "X",
		title: str = "1D Signal"
	) -> S1D:
		"""
		Create `S1D` instance from basic data structures.

		.. code-block:: python

			import modusa as ms
			x = ms.s1d.from_array([1, 2, 3])
			print(x)
			x.print_info()
		
		Parameters
		----------
		y: array-like
			- Data values.
		x: array-like | None
			- Corresponding axis values.
			- Default: None → Creates an integer indexing.
		y_label: str
			- Y label for the signal.
			- Default: "Y"
		x_label: str
			- X label for the signal.
			- Default: "X"
		title: str
			- Title for the signal.
			- Default: "1D Signal"
		Returns
		-------
		S1D
			An instance of S1D.
		"""
		assert isinstance(y, (np.ndarray, list, float, int, np.generic))
		assert isinstance(x, (np.ndarray, list, float, int, np.generic)) or x is None
		assert isinstance(y_label, str) and isinstance(x_label, str) and isinstance(title, str)
		
		if isinstance(y, (float, int, np.generic)): y = [y] # Convert to list of 1 element
		if isinstance(x, (float, int, np.generic)): x = [x] # Convert to list of 1 element
		
		y = np.asarray(y)
		if x is None: x = np.arange(y.shape[0])
		else: x = np.asarray(x)
		
		assert y.ndim ==1 and x.ndim == 1, "S1D must have only one dimension"
		assert y.shape == x.shape, "Shape mismatch"
		
		y = Data(values=y, label=y_label)
		x = SAx(values=x, label=x_label) # Creating a signal axis instance
		
		return S1D(y=y, x=x, title=title)
	
	@classmethod
	def zeros(cls, shape) -> S1D:
		"""
		Create `S1D` instance with all zeros.

		.. code-block:: python

			import modusa as ms
			y = ms.s1d.zeros(10)
			print(y)
			y.print_info()
		
		Parameters
		----------
		shape: int | tuple[int]
			- Shape of the signal with zeros.
			- Must be 1 dimensional
			- E.g. 10 or (10, )
		Returns
		-------
		S1D
			An instance of S1D.
		"""
		assert isinstance(shape, (int, tuple))
		y = np.zeros(shape)
		
		return cls.from_array(y=y, title="Zeros")
	
	@classmethod
	def zeros_like(cls, signal: S1D) -> S1D:
		"""
		Create `S1D` instance similar to `signal`
		but with all entries being zeros.

		.. code-block:: python

			import modusa as ms
			signal = ms.s1d.from_array([1, 2, 3])
			y = ms.s1d.zeros_like(signal)
			print(y)
			x.print_info()
		
		Parameters
		----------
		signal: S1D
			- Reference signal to create zeros like that.
		Returns
		-------
		S1D
			An instance of S1D.
		"""
		
		assert signal.__class__ in [S1D]
		
		y = np.zeros(signal.shape)
		y_label = signal._y_label
		x = signal._x
		x_label = signal._x_label
		title = signal._title
		
		return cls.from_array(y=y, x=x, y_label=y_label, x_label=x_label, title=title)
	
	
	@classmethod
	def ones(cls, shape: int | tuple[int, int]) -> S1D:
		"""
		Create `S1D` instance with all ones.

		.. code-block:: python

			import modusa as ms
			y = ms.s1d.ones(10)
			print(y)
			y.print_info()
		
		Parameters
		----------
		shape: int | tuple[int]
			- Shape of the signal with ones.
			- Must be 1 dimensional
			- E.g. 10 or (10, )
		Returns
		-------
		S1D
			An instance of S1D.
		"""
		assert isinstance(shape, (int, tuple))
		y = np.ones(shape)
		
		return cls.from_array(y=y, title="Ones")
	
	@classmethod
	def ones_like(cls, signal: S1D) -> S1D:
		"""
		Create `S1D` instance similar to `signal`
		but with all entries being ones.

		.. code-block:: python

			import modusa as ms
			signal = ms.s1d.from_array([1, 2, 3])
			y = ms.s1d.ones_like(signal)
			print(y)
			y.print_info()
		
		Parameters
		----------
		signal: S1D
			- Reference signal to create ones like that.
		Returns
		-------
		S1D
			An instance of S1D.
		"""
		
		assert signal.__class__ in [S1D]
		
		y = np.ones(signal.shape)
		y_label = signal._y_label
		x = signal._x
		x_label = signal._x_label
		title = signal._title
		
		return cls.from_array(y=y, x=x, y_label=y_label, x_label=x_label, title=title)
	
	@classmethod
	def random(cls, shape: int) -> S1D:
		"""
		Create `S1D` instance with random entries.

		.. code-block:: python

			import modusa as ms
			y = ms.s1d.random(10)
			print(y)
			y.print_info()
		
		Parameters
		----------
		shape: int | tuple[int]
			- Shape of the signal.
			- Must be 1 dimensional
			- E.g. 10 or (10, )
		Returns
		-------
		S1D
			An instance of S1D with random values.
		"""
		assert isinstance(shape, (int, tuple))
		y = np.random.random(shape)
		
		return cls.from_array(y=y, title="Random")
	
	@classmethod
	def random_like(cls, signal: S1D) -> S1D:
		"""
		Create `S1D` instance similar to `signal`
		but with all entries being ones.

		.. code-block:: python

			import modusa as ms
			signal = ms.s1d.from_array([1, 2, 3])
			y = ms.s1d.random_like(signal)
			print(y)
			y.print_info()
		
		Parameters
		----------
		signal: S1D
			- Reference signal to create one with random entries.
		Returns
		-------
		S1D
			An instance of S1D with random values.
		"""
		
		assert signal.__class__ in [S1D]
		
		y = np.random.random(signal.shape)
		y_label = signal._y_label
		x = signal._x
		x_label = signal._x_label
		title = signal._title
		
		return cls.from_array(y=y, x=x, y_label=y_label, x_label=x_label, title=title)
	


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from .base import ModusaGenerator
from modusa.models.s2d import S2D
from modusa.models.s_ax import SAx
from modusa.models.data import Data
import numpy as np

class S2DGen(ModusaGenerator):
	"""
	Provides user friendly APIs to generate instances of different `S2D`
	instances.
	"""
	
	#--------Meta Information----------
	_name = "S2DGeneratator"
	_description = "APIs to generate instances of different `S2D` instances"
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-27"
	#----------------------------------
	
	@staticmethod
	def from_array(
		M: np.ndarray | list | float | int | np.generic,
		y: np.ndarray | list | float | int | np.generic | None = None,
		x: np.ndarray | list | float | int | np.generic | None = None,
		M_label: str = "M",
		y_label: str = "Y",
		x_label: str = "X",
		title: str = "2D Signal"
	) -> S2D:
		"""
		Create `S2D` instance from basic data structures.

		.. code-block:: python

			import modusa as ms
			M = ms.s1d.from_array([1, 2, 3])
			print(M)
			M.print_info()
		
		Parameters
		----------
		M: np.ndarray | list | float | int | np.generic
			- Data values.
		y: np.ndarray | list | float | int | np.generic
			- y axis values.
			- Default: None → Creates an integer indexing.
		x: np.ndarray | list | float | int | np.generic | None
			- x axis values.
			- Default: None → Creates an integer indexing.
		M_label: str
			- Label for the data.
			- Default: "M"
		y_label: str
			- Y label for the signal.
			- Default: "Y"
		x_label: str
			- X label for the signal.
			- Default: "X"
		title: str
			- Title for the signal.
			- Default: "2D Signal"
		Returns
		-------
		S2D
			An instance of S2D.
		"""
		assert isinstance(M, (np.ndarray, list, float, int, np.generic))
		assert isinstance(x, (np.ndarray, list, float, int, np.generic)) or x is None
		assert isinstance(y, (np.ndarray, list, float, int, np.generic)) or y is None
		assert isinstance(M_label, str) and isinstance(y_label, str) and isinstance(x_label, str) and isinstance(title, str)
		
		if isinstance(M, (float, int, np.generic)): M = [[M]] # Convert to list of 1 element
		if isinstance(y, (float, int, np.generic)): y = [y] # Convert to list of 1 element
		if isinstance(x, (float, int, np.generic)): x = [x] # Convert to list of 1 element
		
		M = np.asarray(M)
		assert M.ndim == 2
		
		if y is None: y = np.arange(M.shape[0])
		else: y = np.asarray(y)
		assert y.ndim == 1
		
		if x is None: x = np.arange(M.shape[1])
		else: x = np.asarray(x)
		assert x.ndim == 1
		
		assert y.shape[0] == M.shape[0], "Shape mismatch"
		assert x.shape[0] == M.shape[1], "Shape mismatch"
		
		y_sax = SAx(values=y, label=y_label) # Creating a signal axis instance
		x_sax = SAx(values=x, label=x_label) # Creating a signal axis instance
		
		M = Data(values=M, label=M_label)
		
		return S2D(M=M, y=y_sax, x=x_sax, title=title)
	
	@classmethod
	def zeros(cls, shape: tuple[int, int]) -> S2D:
		"""
		Create `S2D` instance with all zeros.

		.. code-block:: python

			import modusa as ms
			M = ms.s2d.zeros((10, 5))
			print(M)
			M.print_info()
		
		Parameters
		----------
		shape: tuple[int, int]
			- Shape of the signal with zeros.
			- Must be 1 dimensional
			- E.g. (10, 5)
		Returns
		-------
		S2D
			An instance of S2D.
		"""
		assert isinstance(shape, tuple)
		M = np.zeros(shape)
		
		return cls.from_array(M=M, title="Zeros")
	
	@classmethod
	def zeros_like(cls, signal: S2D) -> S2D:
		"""
		Create `S2D` instance similar to `signal`
		but with all entries being zeros.

		.. code-block:: python

			import modusa as ms
			signal = ms.s2d.from_array([[1, 2, 3], [4, 5, 6]])
			M = ms.s2d.zeros_like(signal)
			print(M)
			M.print_info()
		
		Parameters
		----------
		signal: S2D
			- Reference signal to create zeros like that.
		Returns
		-------
		S2D
			An instance of S2D.
		"""
		
		assert signal.__class__ in [S2D]
		
		M = np.zeros(signal.shape)
		y = signal._y
		x = signal._x
		
		M_label = signal._M_label
		y_label = signal._y_label
		x_label = signal._x_label
		title = signal._title
		
		return cls.from_array(M=M, y=y, x=x, M_label=M_label, y_label=y_label, x_label=x_label, title=title)
	
	
	@classmethod
	def ones(cls, shape: tuple[int, int]) -> S2D:
		"""
		Create `S2D` instance with all ones.

		.. code-block:: python

			import modusa as ms
			M = ms.s2d.ones((10, 5))
			print(M)
			M.print_info()
		
		Parameters
		----------
		shape: tuple[int, int]
			- Shape of the signal with ones.
			- Must be 1 dimensional
			- E.g. (10, 5)
		Returns
		-------
		S2D
			An instance of S2D.
		"""
		assert isinstance(shape, tuple)
		M = np.ones(shape)
		
		return cls.from_array(M=M, title="Ones")
	
	@classmethod
	def ones_like(cls, signal: S2D) -> S2D:
		"""
		Create `S2D` instance similar to `signal`
		but with all entries being ones.

		.. code-block:: python

			import modusa as ms
			signal = ms.s2d.from_array([[1, 2, 3], [4, 5, 6]])
			M = ms.s2d.ones_like(signal)
			print(M)
			M.print_info()
		
		Parameters
		----------
		signal: S2D
			- Reference signal to create ones like that.
		Returns
		-------
		S2D
			An instance of S2D.
		"""
		
		assert signal.__class__ in [S2D]
		
		M = np.ones(signal.shape)
		y = signal._y
		x = signal._x
		
		M_label = signal._M_label
		y_label = signal._y_label
		x_label = signal._x_label
		title = signal._title
		
		return cls.from_array(M=M, y=y, x=x, M_label=M_label, y_label=y_label, x_label=x_label, title=title)
	
	@classmethod
	def random(cls, shape: tuple[int, int]) -> S2D:
		"""
		Create `S2D` instance with random entries.

		.. code-block:: python

			import modusa as ms
			y = ms.s2d.random((10, 5))
			print(y)
			y.print_info()
		
		Parameters
		----------
		shape: tuple[int, int]
			- Shape of the signal.
			- Must be 1 dimensional
			- E.g. (10, 5)
		Returns
		-------
		S2D
			An instance of S2D with random values.
		"""
		assert isinstance(shape, tuple)
		M = np.random.random(shape)
		
		return cls.from_array(M=M, title="Random")
	
	@classmethod
	def random_like(cls, signal: S2D) -> S2D:
		"""
		Create `S2D` instance similar to `signal`
		but with all entries being ones.

		.. code-block:: python

			import modusa as ms
			signal = ms.s2d.from_array([[1, 2, 3], [4, 5, 6]])
			M = ms.s2d.random_like(signal)
			print(M)
			M.print_info()
		
		Parameters
		----------
		signal: S2D
			- Reference signal.
		Returns
		-------
		S2D
			An instance of S2D with random values.
		"""
		
		assert signal.__class__ in [S2D]
		
		M = np.random.random(signal.shape)
		y = signal._y
		x = signal._x
		
		M_label = signal._M_label
		y_label = signal._y_label
		x_label = signal._x_label
		title = signal._title
		
		return cls.from_array(M=M, y=y, x=x, M_label=M_label, y_label=y_label, x_label=x_label, title=title)
	
	
	


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from modusa.tools.base import ModusaTool
from typing import Any
from pathlib import Path
import yt_dlp


def download(url, content_type, output_dir):
	"""
	Downloads audio/video from YouTube.

	.. code-block:: python
		
		# To download audio
		import modusa as ms
		audio_fp = ms.download(
			url="https://www.youtube.com/watch?v=lIpw9-Y_N0g", 
			content_type="audio", 
			output_dir=".")
	
	Parameters
	----------
	url: str
		Link to the YouTube video.
	content_type: str
		"audio" or "video"
	output_dir: str | Path
		Directory to save the YouTube content.
	
	Returns
	-------
	Path
		File path of the downloaded content.
	
	"""
	if content_type == "audio":
		output_dir = Path(output_dir)
		output_dir.mkdir(parents=True, exist_ok=True)
		
		ydl_opts = {
				'format': 'bestaudio/best',
				'outtmpl': f'{output_dir}/%(title)s.%(ext)s',
				'quiet': True,
			}
		
		with yt_dlp.YoutubeDL(ydl_opts) as ydl:
			info = ydl.extract_info(url, download=True)
			return Path(info['requested_downloads'][0]['filepath'])
		
	elif content_type == "video":
		output_dir = Path(output_dir)
		output_dir.mkdir(parents=True, exist_ok=True)
		
		ydl_opts = {
			'format': 'bestvideo+bestaudio/best',  # High quality
			'outtmpl': str(output_dir / '%(title)s.%(ext)s'),
			'merge_output_format': 'mp4',
			'quiet': True,  # Hide verbose output
		}
		
		with yt_dlp.YoutubeDL(ydl_opts) as ydl:
			info = ydl.extract_info(url, download=True)
			return Path(info['requested_downloads'][0]['filepath'])
	else:
		raise excp.InputValueError(f"`content_type` can either take 'audio' or 'video' not {content_type}")

	
	


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from .base import ModusaGenerator
from modusa.models.tds import TDS
from modusa.models.t_ax import TAx
from modusa.models.data import Data
import numpy as np

class TDSGen(ModusaGenerator):
	"""
	Provides user friendly APIs to generate instances of different 
	`TDS` instances.
	"""
	
	#--------Meta Information----------
	_name = "TDSGenerator"
	_description = ""
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-27"
	#----------------------------------
	
	@staticmethod
	def from_array(
		y: np.ndarray | list | float | int | np.generic,
		sr: float | int = 1.0,
		t0: float | int = 0.0,
		y_label: str = "Y",
		t_label: str = "Time (sec)",
		title: str = "Time Domain Signal"
	) -> TDS:
		"""
		Create `TDS` instance from basic data structures.

		.. code-block:: python

			import modusa as ms
			t = ms.tds.from_array([1, 2, 3])
			print(t)
			t.print_info()
		
		Parameters
		----------
		y: np.ndarray | list | float | int | np.generic
			- Data values.
		sr: float | int
			- Sampling rate.
		t0: float | int
			- Start timestamp.
		y_label: str
			- Y label for the signal.
			- Default: "Y"
		t_label: str
			- T label for the signal.
			- Default: "Time (sec)"
		title: str
			- Title for the signal.
			- Default: "1D Signal"
		Returns
		-------
		TDS
			An instance of TDS.
		"""
		assert isinstance(y, (np.ndarray, list, float, int, np.generic))
		assert isinstance(sr, (int, float)) and isinstance(t0, (int, float))
		assert isinstance(y_label, str) and isinstance(t_label, str) and isinstance(title, str)
		
		if isinstance(y, (float, int, np.generic)): y = [y] # Convert to list of 1 element
		y = np.asarray(y)
		assert y.ndim == 1
		
		sr = float(sr)
		t0 = float(t0)
		
		y = Data(values=y, label=y_label)
		t = TAx(n_points=y.shape[0], sr=sr, t0=t0, label=t_label) # Creating a signal axis instance
		
		return TDS(y=y, t=t, title=title)
	
	@classmethod
	def zeros(cls, shape, sr=1.0, t0=0.0) -> TDS:
		"""
		Create `TDS` instance with all zeros.

		.. code-block:: python

			import modusa as ms
			y = ms.tds.zeros(10, sr=10)
			print(y)
			y.print_info()
		
		Parameters
		----------
		shape: int | tuple[int]
			- Shape of the signal with zeros.
			- Must be 1 dimensional
			- E.g. 10 or (10, )
		sr: float | int
			- Sampling rate.
		t0: float | int
			- Start timestamp.
		Returns
		-------
		TDS
			An instance of TDS.
		"""
		assert isinstance(shape, (int, tuple))
		y = np.zeros(shape)
		
		return cls.from_array(y=y, sr=sr, t0=t0, title="Zeros")
	
	@classmethod
	def zeros_like(cls, signal, shape=None) -> TDS:
		"""
		Create `TDS` instance similar to `signal`
		but with all entries being zeros.

		.. code-block:: python

			import modusa as ms
			signal = ms.tds.from_array([1, 2, 3])
			y = ms.tds.zeros_like(signal)
			print(y)
			x.print_info()
		
		Parameters
		----------
		signal: TDS
			- Reference signal to create zeros like that.
		Returns
		-------
		TDS
			An instance of TDS.
		"""
		
		assert isinstance(signal, TDS)
		
		shape = signal.shape if shape is None else shape
		
		return cls.from_array(y=np.zeros(shape), sr=signal.t.sr, t0=signal.t.t0, y_label=signal.y.label, t_label=signal.t.label, title=signal.title)
	
	
	@classmethod
	def ones(cls, shape, sr=1.0, t0=0.0) -> TDS:
		"""
		Create `TDS` instance with all ones.

		.. code-block:: python

			import modusa as ms
			y = ms.tds.ones(10)
			print(y)
			y.print_info()
		
		Parameters
		----------
		shape: int | tuple[int]
			- Shape of the signal with ones.
			- Must be 1 dimensional
			- E.g. 10 or (10, )
		sr: float | int
			- Sampling rate.
		t0: float | int
			- Start timestamp.
		Returns
		-------
		TDS
			An instance of TDS.
		"""
		assert isinstance(shape, (int, tuple))
		y = np.ones(shape)
		
		return cls.from_array(y=y, sr=sr, t0=t0, title="Ones")
	
	@classmethod
	def ones_like(cls, signal, shape=None) -> TDS:
		"""
		Create `TDS` instance similar to `signal`
		but with all entries being ones.
		
		.. code-block:: python

			import modusa as ms
			signal = ms.tds.from_array([1, 2, 3])
			y = ms.tds.ones_like(signal)
			print(y)
			y.print_info()
		
		Parameters
		----------
		signal: TDS
			- Reference signal to create ones like that.
		Returns
		-------
		TDS
			An instance of TDS.
		"""
		assert isinstance(signal, TDS)
		
		shape = signal.shape if shape is None else shape
		
		return cls.from_array(y=np.ones(shape), sr=signal.t.sr, t0=signal.t.t0, y_label=signal.y.label, t_label=signal.t.label, title=signal.title)
	
	@classmethod
	def random(cls, shape, sr=1.0, t0=0.0) -> TDS:
		"""
		Create `TDS` instance with random entries.

		.. code-block:: python

			import modusa as ms
			x = ms.tds.random(10)
			print(x)
			x.print_info()
		
		Parameters
		----------
		shape: int | tuple[int]
			- Shape of the signal.
			- Must be 1 dimensional
			- E.g. 10 or (10, )
		sr: float | int
			- Sampling rate.
		t0: float | int
			- Start timestamp.
		Returns
		-------
		TDS
			An instance of TDS with random values.
		"""
		assert isinstance(shape, (int, tuple))
		y = np.random.random(shape)
		
		return cls.from_array(y=y, sr=sr, t0=t0, title="Random")
	
	@classmethod
	def random_like(cls, signal, shape=None) -> TDS:
		"""
		Create `TDS` instance similar to `signal`
		but with all entries being ones.

		.. code-block:: python

			import modusa as ms
			signal = ms.tds.from_array([1, 2, 3])
			y = ms.tds.random_like(signal)
			print(y)
			y.print_info()
		
		Parameters
		----------
		signal: TDS
			- Reference signal to create one with random entries.
		Returns
		-------
		TDS
			An instance of TDS with random values.
		"""
		assert isinstance(signal, TDS)
		
		shape = signal.shape if shape is None else shape
		
		return cls.from_array(y=np.random.random(shape), sr=signal.t.sr, t0=signal.t.t0, y_label=signal.y.label, t_label=signal.t.label, title=signal.title)
	


#!/usr/bin/env python3

from modusa.models.s1d import S1D
from modusa.models.s2d import S2D
import matplotlib.pyplot as plt
import numpy as np
from collections import defaultdict
import itertools

def _in_notebook() -> bool:
	"""
	To check if we are in jupyter notebook environment.
	"""
	try:
		from IPython import get_ipython
		shell = get_ipython()
		return shell and shell.__class__.__name__ == "ZMQInteractiveShell"
	except ImportError:
		return False

def plot_multiple_signals(
	*args,
	loc = None,
	x_lim: tuple[float, float] | None = None,
	highlight_regions: list[tuple[float, float, str]] | None = None,
	vlines: list[float, ...] | None = None,
) -> plt.Figure:
	"""
	Plots multiple instances of uniform `Signal1D` and `Signal2D`
	with proper formatting and time aligned.

	Parameters
	----------
	loc: tuple[int]
		- The len should match the number of signals sent as args.
		- e.g. (0, 1, 1) => First plot at ax 0, second and third plot at ax 1
		- Default: None => (0, 1, 2, ...) all plots on a new ax.
	highlight_regions: list[tuple[float, float, str]]
		-
	
	"""
	
	assert len(args) >= 1, "No signal provided to plot"
	signals = args
	
	for signal in signals: 
		if not isinstance(signal, (S1D, S2D)):
			raise TypeError(f"Invalid signal type {type(signal)}")
	
	if loc is None: # => We will plot all the signals on different subplots
		loc = tuple([i for i in range(len(args))]) # Create (0, 1, 2, ...) that serves as ax number for plots
	else: # Incase loc is provided, we make sure that we do it for each of the signals
		assert len(args) == len(loc)
	
	# Make sure that all the elements in loc do not miss any number in between (0, 1, 1, 2, 4) -> Not allowed
	assert min(loc) == 0
	max_loc = max(loc)
	for i in range(max_loc):
		if i not in loc:
			raise ValueError(f"Invalid `loc` values, it should not have any missing integer in between.")
	
	# Create a dict that maps subplot to signals that need to be plotted on that subplot e.g. {0: [signal1, signal3], ...}
	subplot_signal_map = defaultdict(list)
	for signal, i in zip(args, loc):
		subplot_signal_map[i].append(signal)
	
	# We need to create a figure with right configurations
	height_ratios = []
	height_1d_subplot = 0.4
	height_2d_subplot = 1
	n_1d_subplots = 0
	n_2d_subplots = 0
	for l, signals in subplot_signal_map.items():

		# If there is any 2D signal, the subplot will be 2D
		if any(isinstance(s, S2D) for s in signals):
			n_2d_subplots += 1
			height_ratios.append(height_2d_subplot)
		
		# If all are 1D signal, the subplot will be 1D
		elif all(isinstance(s, S1D) for s in signals):
			n_1d_subplots += 1
			height_ratios.append(height_1d_subplot)
	
	
	n_subplots = n_1d_subplots + n_2d_subplots
	fig_width = 15
	fig_height = n_1d_subplots * 2 + n_2d_subplots * 4 # This is as per the figsize height set in the plotter tool
	fig, axs = plt.subplots(n_subplots, 2, figsize=(fig_width, fig_height), width_ratios=[1, 0.01], height_ratios=height_ratios) # 2nd column for cbar
	
	if n_subplots == 1:
		axs = [axs]  # axs becomes list of one pair [ (ax, cbar_ax) ]
	
	# We find the x axis limits as per the max limit for all the signals combined, so that all the signals can be seen.
	if x_lim is None:
		x_min = min(np.min(signal.x.values) for signal in args)
		x_max = max(np.max(signal.x.values) for signal in args)
		x_lim = (x_min, x_max)
		
	for l, signals in subplot_signal_map.items():
		# Incase we have plot multiple signals in the same subplot, we change the color
		fmt_cycle = itertools.cycle(['k-', 'r-', 'g-', 'b-', 'm-', 'c-', 'y-'])
		
		# For each subplot, we want to know if it is 2D or 1D
		if any(isinstance(s, S2D) for s in signals): is_1d_subplot = False
		else: is_1d_subplot = True
		
		if is_1d_subplot: # All the signals are 1D
			for signal in signals:
				fmt = next(fmt_cycle)
				if len(signals) == 1: # highlight region works properly only if there is one signal for a subplot
					signal.plot(axs[l][0], x_lim=x_lim, highlight_regions=highlight_regions, show_grid=True, vlines=vlines, fmt=fmt, legend=signal._title)
				else:
					y, x = signal._x._values, signal._y._values
					signal.plot(axs[l][0], x_lim=x_lim, show_grid=True, vlines=vlines, fmt=fmt, legend=signal.title, y_label=signal.y.label, x_label=signal.x.label, title="")
				
			# Remove the colorbar column (if the subplot is 1d)
			axs[l][1].remove()
			
		if not is_1d_subplot: # Atleast 1 signal is 2D, we we have a 2D subplot
			for signal in signals:
				if len(signals) == 1: # Only one 2D signal is to be plotted
					signal.plot(axs[l][0], x_lim=x_lim, show_colorbar=True, cax=axs[l][1], highlight_regions=highlight_regions, vlines=vlines)
				else:
					if isinstance(signal, S1D):
						fmt = next(fmt_cycle)
						signal.plot(axs[l][0], x_lim=x_lim, show_grid=True, vlines=vlines, fmt=fmt, legend=signal.title, y_label=signal.y.label, x_label=signal.x.label, title="")
					elif isinstance(signal, S2D):
						signal.plot(axs[l][0], x_lim=x_lim, show_colorbar=True, cax=axs[l][1], vlines=vlines, x_label=signal.x.label, y_label=signal.y.label, title="")
		
		# We set the xlim, this will align all the signals automatically as they are on the same row
		for l in range(n_subplots):
			axs[l][0].set_xlim(x_lim)
		
	if _in_notebook():
		plt.tight_layout()
		plt.close(fig)
		return fig
	else:
		plt.tight_layout()
		plt.show()
		return fig


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .base import ModusaSignalAxis
from .data import Data
from modusa.tools.math_ops import MathOps
from typing import Self, Any, Callable
import numpy as np
import matplotlib.pyplot as plt

class SAx(ModusaSignalAxis):
	"""
	Space to represent any signal's axis.

	Parameters
	----------
	values: array-like
		- Any array-like object on which np.asarray can be applied.
	label: str
		- Label for the axis.
		- Default: None => ''

	Note
	----
	- Use :class:`~modusa.generators.s_ax.SAxGen` API to instantiate this class.
	- It can be uniform/non-uniform.
	- It is numpy compatible, so you can use numpy methods directly on this class object.
	- Since the object of this class represents an axis, any mathematical operations on it will result in another object of :class:`~modusa.models.s1d.S1D` class with `y` being the result of the operation and `x` being the axis itself.
	"""
	
	#--------Meta Information----------
	_name = "Signal Axis"
	_nickname = "axis" # This is to be used in repr/str methods
	_description = "Axis for different types of signals."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-20"
	#----------------------------------
	
	@validate_args_type()
	def __init__(self, values, label = None):
		
		super().__init__() # Instantiating `ModusaSignalAxis` class

		values = np.asarray(values)
		if values.ndim == 0:
			values = np.asarray([values])
		
		assert values.ndim == 1
	
		self._values = values
		self._label = label or ""
	
	#-----------------------------------
	# Properties (User Facing)
	#-----------------------------------
	
	@property
	def values(self) -> np.ndarray:
		return self._values
	
	@property
	def label(self) -> str:
		return self._label
		
	@property
	def shape(self) -> tuple:
		return self.values.shape
	
	@property
	def ndim(self) -> int:
		return self.values.ndim # Should be 1
	
	@property
	def size(self) -> int:
		return self.values.size
	
	#====================================
	
	#------------------------------------
	# Utility methods
	#------------------------------------
	
	def is_same_as(self, other) -> bool:
		"""
		Compare it with another SAx object.

		Parameters
		----------
		other: SAx
			Another object to compare with.
		
		Returns
		-------
		bool
			True if same ow False

		Note
		----
		- We check the shape and all the values.
		- We are not checking the labels for now.
		"""
		
		if other.size == 1: # Meaning it is scalar
			return True
		
		axis1_arr = np.asarray(self)
		axis2_arr = np.asarray(other)
		
		if not isinstance(axis2_arr, type(axis1_arr)):
			return False
		if axis1_arr.shape != axis2_arr.shape:
			return False
		if not np.allclose(axis1_arr, axis2_arr):
			return False
		
		return True
	
	def copy(self) -> Self:
		"""
		Return a new copy of SAx object.
		
		Returns
		-------
		SAx
			A new copy of the SAx object.
		"""
		copied_values = np.asarray(self).copy()
		copied_label = self.label
		
		return self.__class__(values=copied_values, label=copied_label)
	
	def set_meta_info(self, label):
		"""
		Set meta info for the axis.
		
		.. code-block:: python
		
			import modusa as ms
			x = ms.sax.linear(100, 10)
			print(x)
			x = x.set_meta_info("My Axis (unit)")
			print(x)

			# I personally prefer setting it inline
			x = ms.sax.linear(100, 10).set_meta_info("My Axis (unit)")
			print(x)
		
		Parameters
		----------
		label: str
			Label for the axis (e.g. "Time (sec)").
		Returns
		-------
		Self
			A new Self instance with new label.
		"""
		
		if label is None:
			return self
		else:
			return self.__class__(values=self.values.copy(), label=label)
		
	
	def index_of(self, value) -> int:
		"""
		Return the index whose value is closest
		to `value`.

		Parameters
		----------
		value: float
			value to find the index of.
		
		Returns
		-------
		int
			Index with value closest to the `value`
		"""
		from .data import Data
		
		idx = np.argmin(np.abs(self.values - value))
		
		return Data(values=idx, label=None)

	#====================================
	
	#-------------------------------
	# NumPy Protocol
	#-------------------------------
	
	def __array__(self, dtype=None) -> np.ndarray:
		return np.asarray(self.values, dtype=dtype)
	
	def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
		"""
		Provides operation support for the universal functions
		on the Data object.
		"""
		from .s1d import S1D
		from .data import Data
		
		raw_inputs = [x.values if isinstance(x, type(self)) else x for x in inputs]
		
		# Call the actual ufunc
		result = getattr(ufunc, method)(*raw_inputs, **kwargs)
		
		if isinstance(result, (np.ndarray, np.generic)):
			y = Data(values=result, label=None)
			x = self
			return S1D(y=y, x=x, title=None)
		else:
			return result
		
	def __array_function__(self, func, types, args, kwargs):
		"""
		Additional numpy function support.
		"""
		from modusa.utils import np_func_cat as nfc
		from .data import Data
		
		if not all(issubclass(t, type(self)) for t in types):
			return NotImplemented
		
		# Not supporting concatenate like operations as axis any random axis can't be concatenated
		if func in nfc.CONCAT_FUNCS:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
			
		# Single signal input expected
		x = args[0]
		x_arr = np.asarray(x)
		result = func(x_arr, **kwargs)
		
		if func in nfc.REDUCTION_FUNCS:
			# If the number of dimensions is reduced
			if result.ndim == 0:
				return Data(values=result, label=None)
			else:
				raise RuntimeError(f"Unexpected result: `result` has more than 0 dimensions, {result.ndim}")
				
		elif func in nfc.X_NEEDS_ADJUSTMENT_FUNCS:
			# You must define logic for adjusting x
			raise NotImplementedError(f"{func.__name__} requires x-axis adjustment logic.")
			
		else:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
	
	#================================
	
	
	#------------------------------------
	# Visualisation
	#------------------------------------
	
	def plot(self, ax = None, fmt = "b-", show_stem=False):
		"""
		Plot the axis values. This is useful to analyse
		the axis if it is linear or follows some other
		trend.

		.. code-block:: python
			
			import modusa as ms
			x = ms.sax.linear(100, 10)
			x.plot()

		Parameters
		----------
		ax: plt.Axes | None
			- Incase, you want to plot it on your defined matplot ax.
			- If not provided, We create a new figure and return that.
		fmt: str
			- Matplotlib fmt for setting different colors and styles for the plot.
			- General fmt is [color][marker][line], e.g. "bo:"
			- Useful while plotting multiple SAx instances on the same plot.

		Returns
		-------
		plt.Figure | None
			- Figure if ax is None.
			- None is ax is not None.
		"""
		from modusa.tools.plotter import Plotter
		
		fig: plt.Figure | None = Plotter.plot_signal(y=self.values, x=np.arange(len(self)), ax=ax, fmt=fmt, title=self.label, y_label=self.label, x_label="Index", show_stem=show_stem)
		
		return fig
	#====================================
	
	#-----------------------------------
	# Indexing
	#-----------------------------------
	
	def __getitem__(self, key) -> Self:
		"""
		Defining how to index SAx instance.

		.. code-block:: python
			
			import modusa as ms
			x = ms.sax.linear(100, 10)
			print(x)
			print(x[10:20])

		Parameters
		----------
		key: int | slice
			What can go inside the square bracket [] for indexing.
		
		Returns
		-------
		SAx:
			Sliced instance of the axis.
		"""
		
		from .s1d import S1D
		
		if not isinstance(key, (int, slice, tuple)):
			raise TypeError(f"Invalid key type {type(key)}")
	
		sliced_values = self.values[key]
		
		# If the number of dimensions is reduced, add back singleton dims
		while np.ndim(sliced_values) < self.ndim:
			sliced_values = np.expand_dims(sliced_values, axis=0)
		
		return self.__class__(values=sliced_values, label=self.label)
	
	
	def __setitem__(self, key, value):
		"""
		Raises error if trying to set values
		of an axis.

		Meaningful axis are not meant to be altered.
		"""
		raise TypeError("Axis do not support item assignment.")

	#===============================
	
	#-------------------------------
	# Basic arithmetic operations
	#-------------------------------
	def __add__(self, other):
		return np.add(self, other) 
	
	def __radd__(self, other):
		return np.add(other, self)
	
	def __sub__(self, other):
		return np.subtract(self, other)
	
	def __rsub__(self, other):
		return np.subtract(other, self)
	
	def __mul__(self, other):
		return np.multiply(self, other) 
	
	def __rmul__(self, other):
		return np.multiply(other, self)
	
	def __truediv__(self, other):
		return np.divide(self, other) 
	
	def __rtruediv__(self, other):
		return np.divide(other, self)
	
	def __floordiv__(self, other):
		return np.floor_divide(self, other) 
	
	def __rfloordiv__(self, other):
		return np.floor_divide(other, self)
	
	def __pow__(self, other):
		return np.power(self, other) 
	
	def __rpow__(self, other):
		return np.power(other, self)
	
	#===============================
	
	
	#-------------------------------
	# Basic comparison operations
	#-------------------------------
	def __eq__(self, other):
		return np.equal(self, other)
	
	def __ne__(self, other):
		return np.not_equal(self, other)
	
	def __lt__(self, other):
		return np.less(self, other)
	
	def __le__(self, other):
		return np.less_equal(self, other)
	
	def __gt__(self, other):
		return np.greater(self, other)
	
	def __ge__(self, other):
		return np.greater_equal(self, other)
	
	#===============================

	#----------------------------------
	# Information
	#----------------------------------
	
	def print_info(self) -> None:
		"""
		Prints info about the SAx instance.
		
		.. code-block:: python
			
			import modusa as ms
			# For SAx
			x = ms.sax.linear(100, 10)
			x.print_info()
			# For TAx
			x = ms.tax.linear(100, 10)
			x.print_info()
		
		Returns
		-------
		None
		"""
		print("-" * 50)
		print("Axis Info")
		print("-" * 50)
		print(f"{'Label':<20}: {self.label}")
		print(f"{'Shape':<20}: {self.shape}")
		# Inheritance chain
		cls_chain = " → ".join(cls.__name__ for cls in reversed(self.__class__.__mro__[:-1]))
		print(f"{'Inheritance':<20}: {cls_chain}")
		print("=" * 50)
	
	def __str__(self):
		arr_str = np.array2string(self.values, separator=", ", threshold=30, edgeitems=3, max_line_width=120)
		return f"{self._nickname}({arr_str})"
	
	def __repr__(self):
		arr_str = np.array2string(self.values, separator=", ", threshold=30, edgeitems=3, max_line_width=120)
		return f"{self._nickname}({arr_str})"
	#===================================
	


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .base import ModusaSignal
from .s_ax import SAx
from .data import Data
from modusa.tools.math_ops import MathOps
from typing import Self, Any, Callable
from types import NoneType
import numpy as np
import matplotlib.pyplot as plt
import copy

class S2D(ModusaSignal):
	"""
	Space to represent 2D signal.

	Note
	----
	- Use :class:`~modusa.generators.s2d.S2DGen` API to instantiate this class.
	- The signal can have uniform/non-uniform axes.

	Parameters
	----------
	M: Data
		- Data object holding the main 2D array.
	y: SAx
		- Y-axis of the signal.
	x: SAx
		- X-axis of the signal.
	title: str
		- What does the signal represent?
		- e.g. "MySignal"
		- This is used as the title while plotting.
	"""
	
	#--------Meta Information----------
	_name = "Signal 2D"
	_nickname = "signal" # This is to be used in repr/str methods
	_description = "Space to represent 2D signal."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-20"
	#----------------------------------
	
	def __init__(self, M, y, x, title = None):
		super().__init__() # Instantiating `ModusaSignal` class
		
		if not (isinstance(M, Data) and isinstance(y, SAx), isinstance(x, SAx)):
			raise TypeError(f"`M` must be `Data` instance, `y` and `x` must be `SAx` instances, got {type(M)}, {type(y)} and {type(x)}")
			
		assert M.ndim == 2
		assert M.shape[0] == y.shape[0], f"M and y shape mismatch"
		assert M.shape[1] == x.shape[0], f"M and x shape mismatch"
		
		# All these are private and we do not expose it to users directly.
		self._M = M
		self._y = y
		self._x = x
		self._title = title or self._name
		
	#--------------------------------------
	# Properties
	#--------------------------------------
		
	@property
	def M(self) -> Data:
		return self._M
	
	@property
	def y(self) -> SAx:
		return self._y
	
	@property
	def x(self) -> SAx:
		return self._x
	
	@property
	def title(self) -> str:
		return self._title
	
	@property
	def shape(self) -> tuple:
		return self.M.shape
	
	@property
	def ndim(self) -> tuple:
		return self.M.ndim # Should be 2
	
	@property
	def size(self) -> int:
		return self.M.size
	
	#===============================
	
	
	#-------------------------------
	# NumPy Protocol
	#-------------------------------
	
	def __array__(self, dtype=None) -> np.ndarray:
		return np.asarray(self.M.values, dtype=dtype)
	
	def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
		"""
		Supports NumPy universal functions on the Signal1D object.
		"""
		from .data import Data  # Ensure this is the same Data class you're using
		from modusa.utils import np_func_cat as nfc
		
		raw_inputs = [
			np.asarray(obj.M) if isinstance(obj, type(self)) else obj
			for obj in inputs
		]
		
		result = getattr(ufunc, method)(*raw_inputs, **kwargs)
		
		result = Data(values=result, label=None)
		y = self.y.copy()
		x = self.x.copy()
		
		if result.shape[0] != y.shape[0] or result.shape[1] != x.shape[0]:
			raise ValueError(f"`{ufunc.__name__}` caused shape mismatch between data and axis, please create a github issue")
			
		return self.__class__(M=result, y=y, x=x, title=self.title)
	
	def __array_function__(self, func, types, args, kwargs):
		"""
		Additional numpy function support for modusa signals.
		Handles reduction and ufunc-like behavior.
		"""
		from .data import Data
		from modusa.utils import np_func_cat as nfc
		
		if not all(issubclass(t, type(self)) for t in types):
			return NotImplemented
		
		if func in nfc.CONCAT_FUNCS:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
			
		signal = args[0]
		result: Data = func(signal.M, **kwargs)
		axis = kwargs.get("axis", None)
		keepdims = kwargs.get("keepdims", None)

		if func in nfc.REDUCTION_FUNCS:
			if keepdims is None or keepdims is True:
				if axis is None: # Both axes collapsed
					dummy_y = SAx(0, label=None)
					dummy_x = SAx(values=0, label=None)
					return self.__class__(M=result, y=dummy_y, x=dummy_x, title=signal.title)
				
				if isinstance(axis, int): # One of the axis collapsed
					if axis == 0:
						dummy_y = SAx(0, label=None)
						return self.__class__(M=result, y=dummy_y, x=signal.x.copy(), title=signal.title)
					elif axis in [1, -1]:
						dummy_x = SAx(values=0, label=None)
						return self.__class__(M=result, y=signal.y.copy(), x=dummy_x, title=signal.title)
					else:
						raise ValueError
			elif keepdims is False:
				if axis is None: # Return Data
					return result
				if axis == 0: # Return S1D
					from .s1d import S1D
					return S1D(y=result, x=signal.x, title=signal.title)
				elif axis in [1, -1]: # Return S1D
					from .s1d import S1D
					return S1D(y=result, x=signal.y, title=signal.title)
				else:
					raise ValueError
				
		elif func in nfc.X_NEEDS_ADJUSTMENT_FUNCS:
			raise NotImplementedError(f"{func.__name__} requires x-axis adjustment logic.")
			
		else:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
			
			
	#================================
			
	#-------------------------------
	# Indexing
	#-------------------------------
			
	def __getitem__(self, key):
		"""
		Return a sliced or indexed view of the data.
		
		Parameters
		----------
		key : int | slice | S2D
			- Index to apply to the values.
		
		Returns
		-------
		S2D
			- Another sliced S2D object 
		"""
		
		if not isinstance(key, (int, slice, tuple)):
			raise TypeError(f"Invalid key type {type(key)}")
		
		# We slice the data
		sliced_M = self.M[key]
		
		if isinstance(key, (int, slice)):
			sliced_y = self.y[key]
			sliced_x = self.x
		
		if isinstance(key, tuple):
			sliced_y = self.y[key[0]]
			sliced_x = self.x[key[1]]
		
		return self.__class__(M=sliced_M, y=sliced_y, x=sliced_x, title=self.title)
	
		
	
	def __setitem__(self, key, value):
		"""
		Set values at the specified index.
	
		Parameters
		----------
		key : int | slice | array-like | boolean array | S1D
			Index to apply to the values.
		value : int | float | array-like
			Value(s) to set.
		"""
		
		self.M[key] = value  # In-place assignment
		
	#===================================
		
	#-------------------------------
	# Basic arithmetic operations
	#-------------------------------
	def __add__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.add(self, other) 
	
	def __radd__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.add(other, self)
	
	def __sub__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.subtract(self, other)
	
	def __rsub__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.subtract(other, self)
	
	def __mul__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.multiply(self, other) 
	
	def __rmul__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.multiply(other, self)
	
	def __truediv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.divide(self, other) 
	
	def __rtruediv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.divide(other, self)
	
	def __floordiv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.floor_divide(self, other) 
	
	def __rfloordiv__(self, other):
		if not self.has_same_axis_as(other):
			raise ValueError("Axes are not aligned for the operation.")
		return np.floor_divide(other, self)
	
	def __pow__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.power(self, other) 
	
	def __rpow__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.power(other, self)
	
	#===============================
	
	
	#-------------------------------
	# Basic comparison operations
	#-------------------------------
	def __eq__(self, other):
		return np.equal(self, other)
	
	def __ne__(self, other):
		return np.not_equal(self, other)
	
	def __lt__(self, other):
		return np.less(self, other)
	
	def __le__(self, other):
		return np.less_equal(self, other)
	
	def __gt__(self, other):
		return np.greater(self, other)
	
	def __ge__(self, other):
		return np.greater_equal(self, other)
	
	#===============================
	
	
	#-----------------------------------
	# Utility Methods
	#-----------------------------------
	
	def unpack(self):
		"""
		Unpacks the object into easy to work
		with data structures.

		Returns
		-------
		(np.ndarray, np.ndarray, np.ndarray)
			- M: Signal data array.
			- y: Signal Y-axis array.
			- x: Signal X-axis array.
		"""
		
		M = self.M.values
		y = self.y.values
		x = self.x.values
		
		return (M, y, x)
	
	def copy(self) -> Self:
		"""
		Returns a new copy of the signal.

		Returns
		-------
		Self
			A new copy of the object.
		"""
		
		copied_M = self.M.copy()
		copied_y = self.y.copy()
		copied_x = self.x.copy()
		title = self.title # Immutable, hence no need to copy
		
		return self.__class__(M=copied_M, y=copied_y, x=copied_x, title=title)
	
	def set_meta_info(self, title = None, M_label = None, y_label = None, x_label = None) -> None:
		"""
		Set meta info about the signals.

		Parameters
		----------
		title: str
			- Title for the signal
			- e.g. "MyTitle"
		M_label: str
			- Label for the data that matrix is holding.
			- e.g. "Intensity (dB)"
		y_label: str
			- Label for the y-axis.
			- e.g. "Frequency (Hz)"
		x_label: str
			- Label for the x-axis.
			- e.g. "Time (sec)"
		Returns
		-------
		S2D
			A new instance with updated meta info.
		"""
		
		M, y, x = self.M, self.y, self.x
		
		M_label = str(M_label) if M_label is not None else M.label
		y_label = str(y_label) if y_label is not None else y.label
		x_label = str(x_label) if x_label is not None else x.label
		title = str(title) if title is not None else self.title
		
		# We create a new copy of the data and axis
		new_M = M.set_meta_info(label=M_label)
		new_y = y.set_meta_info(label=y_label)
		new_x = x.set_meta_info(label=x_label)
		
		return self.__class__(M=new_M, y=new_y, x=new_x, title=title)
	
	
	def is_same_as(self, other: Self) -> bool:
		"""
		Check if two `S1D` instances are equal.
		"""
		
		if not isinstance(other, type(self)):
			return False
		
		if not self.M.is_same_as(other.M):
			return False
		
		if not self.y.is_same_as(other.y):
			return False
		
		if not self.x.is_same_as(other.x):
			return False
		
		return True
	
	def has_same_axis_as(self, other) -> bool:
		"""
		Check if two 'S1D' instances have same
		axis. Many operations need to satify this.
		"""
		return self.y.is_same_as(other.y) and self.x.is_same_as(other.x)
	
	
	def mask(self, condition, set_to=None) -> Self:
		"""
		Mask the signal based on condition and
		the values can be set.
		
		Parameters
		----------
		condition: Callable
			- Condition function to apply on values of the signal.
			- E.g. lambda x: x > 10
		set_to: Number
			- Number to replace the masked position values.

		Returns
		-------
		S2D
			Masked Signal
		"""
		
		mask = condition(self)
		new_val = set_to
		
		if set_to is None: # Return the mask as the same signal but with booleans
			return mask
	
		else:
			# We apply the mask and update the signal data
			new_data = self.M.mask(condition=condition, set_to=new_val)
			
			# Since we're just updating the data, there is no change in the axis
			return self.__class__(M=new_data, y=self.y.copy(), x=self.x.copy(), title=self.title)
	
	#===================================
	
	
	
		
		
		
	#-----------------------------------
	# Visualisation
	#-----------------------------------
	
	def plot(
		self,
		ax = None,
		cmap = "gray_r",
		title = None,
		M_label = None,
		x_label = None,
		y_label = None,
		y_lim = None,
		x_lim = None,
		highlight_regions = None,
		vlines = None,
		hlines = None,
		origin = "lower",  # or "lower"
		gamma = None,
		show_colorbar = True,
		cax = None,
		show_grid = True,
		tick_mode = "center",  # "center" or "edge"
		n_ticks = None,
	) -> "plt.Figure":
		"""
		Plot the S2D instance using Matplotlib.
	
	
		Parameters
		----------
		ax : matplotlib.axes.Axes | None
			- If you want to plot the signal on a given matplotlib axes, you can pass the ax here. We do not return any figure in this case.
			- If not passed, we create a new figure, plots the signal on it and then return the figure.
		cmap : str, default "gray_r"
			Colormap used for the image.
		title : str | None
			- Title for the plot.
			- If not passed, we use the default set during signal instantiation.
		y_lim : tuple[float, float] | None
			- Limits for the y-axis.
		x_lim : tuple[float, float] | None
			- Limits for the x-axis.
		vlines: list[float]
			- List of x values to draw vertical lines.
			- e.g. [10, 13.5]
		hlines: list[float]
			- List of data values to draw horizontal lines.
			- e.g. [10, 13.5]
		highlight_regions : list[tuple[float, float, str]] | None
			- List of time intervals to highlight on the plot.
			- [(start, end, 'tag')]
		origin : {"lower", "upper"}, default "lower"
			Origin position for the image (for flipping vertical axis).
		gamma : float or int, optional
			If specified, apply log-compression using log(1 + S * factor).
		show_colorbar : bool
			- Whether to display the colorbar.
			- Defaults to True.
		cax : matplotlib.axes.Axes | None
			- Axis to draw the colorbar on. If None, uses default placement.
			- Defaults to None
		show_grid : bool
			- Whether to show the major gridlines.
			- Defaults to None.
		tick_mode : {"center", "edge"}
			- Whether to place ticks at bin centers or edges.
			- Default to "center"
		n_ticks : tuple[int]
			- Number of ticks (y_ticks, x_ticks) to display on each axis.
			- Defaults to None
	
		Returns
		-------
		matplotlib.figure.Figure | None
			- The figure object containing the plot.
			- None if ax is provided.
		"""
		from modusa.tools.plotter import Plotter
		
		M, y, x = self._M, self._y, self._x
		M_val, y_val, x_val = M._values, y._values, x._values
		
		M_label = M_label or M._label
		y_label = y_label or y._label
		x_label = x_label or x._label
		
		title = title or self._title
		
		fig = Plotter.plot_matrix(M=M_val, r=y_val, c=x_val, ax=ax, cmap=cmap, title=title, M_label=M_label, r_label=y_label, c_label=x_label, r_lim=y_lim, c_lim=x_lim,
		highlight_regions=highlight_regions, vlines=vlines, hlines=hlines, origin=origin, gamma=gamma, show_colorbar=show_colorbar, cax=cax, show_grid=show_grid,
		tick_mode=tick_mode, n_ticks=n_ticks)
		
		return fig
		
	#===================================
	
	#-----------------------------------
	# Information
	#-----------------------------------
	
	def print_info(self) -> None:
		"""Print key information about the signal."""
		
		print("-"*50)
		print(f"{'Title':<20}: {self._title}")
		print("-"*50)
		print(f"{'Type':<20}: {self.__class__.__name__}")
		print(f"{'Shape':<20}: {self.shape} (freq bins × time frames)")
		
		# Inheritance chain
		cls_chain = " → ".join(cls.__name__ for cls in reversed(self.__class__.__mro__[:-1]))
		print(f"{'Inheritance':<20}: {cls_chain}")
		print("=" * 50)
		
	def __str__(self):
		arr_str = np.array2string(
			np.asarray(self),
			separator=", ",
			threshold=20,       # limit number of elements shown
			edgeitems=2,          # show first/last 2 rows and columns
			max_line_width=120,   # avoid wrapping
		)
		return f"{self._nickname}({arr_str})"
	
	def __repr__(self):
		arr_str = np.array2string(
			np.asarray(self),
			separator=", ",
			threshold=20,       # limit number of elements shown
			edgeitems=2,          # show first/last 2 rows and columns
			max_line_width=120,   # avoid wrapping
		)
		return f"{self._nickname}({arr_str})"
	
	#===================================


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .base import ModusaSignal
from .data import Data
from .s_ax import SAx
from .t_ax import TAx
from modusa.tools.math_ops import MathOps
from typing import Self, Any, Callable
from types import NoneType
import numpy as np
import matplotlib.pyplot as plt
import copy

class S1D(ModusaSignal):
	"""
	Space to represent any 1D Signal.
	
	Note
	----
	- Use :class:`~modusa.generators.s1d.S1DGen` API to instantiate this class.
	- The signal can have uniform/non-uniform axis.


	Parameters
	----------
	data: Data
		- Data object holding the main array.
	sax: SAx
		- Axis for the signal.
	title: str
		- Title for the signal.
		- Default: None => ''
		- This is used as the title while plotting.
	"""
	
	#--------Meta Information----------
	_name = "Signal 1D"
	_nickname = "signal" # This is to be used in repr/str methods
	_description = "Space to represent any 1D Signal."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-20"
	#----------------------------------
	
	def __init__(self, y, x, title = None):
		super().__init__() # Instantiating `ModusaSignal` class
		
		if not (isinstance(y, Data) and isinstance(x, SAx)):
			raise TypeError(f"`y` must be `Data` instance and `x` must be `SAx` instance, got {type(y)} and {type(x)}")
		
		assert y.ndim == 1
		assert y.shape == x.shape, f"y and x shape must match"
		
		# All these are private and we do not expose it to users directly.
		self._y = y
		self._x = x
		self._title = title or self._name
		
	
	#-----------------------------------
	# Properties (User Facing)
	#-----------------------------------
	
	@property
	def y(self) -> Data:
		return self._y
	
	@property
	def x(self) -> SAx:
		return self._x
	
	@property
	def title(self) -> str:
		return self._title
	
	@property
	def shape(self) -> tuple:
		return self.y.shape
	
	@property
	def ndim(self) -> int:
		return self.y.ndim # Should be 1
	
	@property
	def size(self) -> int:
		return self.y.size
	
	#===================================
	
	#-------------------------------
	# NumPy Protocol
	#-------------------------------
	
	def __array__(self, dtype=None) -> np.ndarray:
		return np.asarray(self.y.values, dtype=dtype)
	
	def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
		"""
		Supports NumPy universal functions on the Signal1D object.
		"""
		from .data import Data  # Ensure this is the same Data class you're using
		from modusa.utils import np_func_cat as nfc
		
		raw_inputs = [
			np.asarray(obj.y) if isinstance(obj, type(self)) else obj
			for obj in inputs
		]
		
		result = getattr(ufunc, method)(*raw_inputs, **kwargs)
		
		y = Data(values=result, label=None)  # label=None or you could copy from self.y.label
		x = self.x.copy()
		
		if y.shape != x.shape:
			raise ValueError(f"`{ufunc.__name__}` caused shape mismatch between data and axis, please create a github issue")
		
		return self.__class__(y=y, x=x, title=self.title)
		
	def __array_function__(self, func, types, args, kwargs):
		"""
		Additional numpy function support.
		"""
		from modusa.utils import np_func_cat as nfc
		
		if not all(issubclass(t, type(self)) for t in types):
			return NotImplemented
		
		# Not supporting concatenate like operations as axis any random axis can't be concatenated
		if func in nfc.CONCAT_FUNCS:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")

		# Single signal input expected
		signal = args[0]
		result: Data = func(signal.y, **kwargs)
		axis = kwargs.get("axis", None)
		keepdims = kwargs.get("keepdims", None)
	
		if func in nfc.REDUCTION_FUNCS:
			if keepdims is None or keepdims is True: # Default state is True => return 1D signal by wrapping the scalar
				from .s_ax import SAx
				dummy_x = SAx(0, "")
				return self.__class__(y=result, x=dummy_x, title=signal.title)
			elif keepdims is False: # Return Data
				from .data import Data
				return Data(values=result, label=None)
	
		elif func in nfc.X_NEEDS_ADJUSTMENT_FUNCS:
			# You must define logic for adjusting x
			raise NotImplementedError(f"{func.__name__} requires x-axis adjustment logic.")
			
		else:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
	
	#================================
		
	#-------------------------------
	# Indexing
	#-------------------------------
	
	def __getitem__(self, key):
		"""
		Return a sliced or indexed view of the data.
		
		Parameters
		----------
		key : int | slice | S1D
			- Index to apply to the values.
		
		Returns
		-------
		S1D
			A new S1D object with sliced values and same meta data.
		"""
		
		if not isinstance(key, (int, slice)):
			raise TypeError(f"Invalid key type: {type(key)}")
		
		sliced_y = self.y[key]
		sliced_x = self.x[key]
		
		return self.__class__(y=sliced_y, x=sliced_x, title=self.title)
	
	def __setitem__(self, key, value):
		"""
		Set values at the specified index.
	
		Parameters
		----------
		key : int | slice | array-like | boolean array | S1D
			Index to apply to the values.
		value : int | float | array-like
			Value(s) to set.
		"""
		
		if not isinstance(key, (int, slice)):
			raise TypeError(f"Invalid key type: {type(key)}")
		
		self.y[key] = value  # In-place assignment
		
	#===================================
		
	#-------------------------------
	# Basic arithmetic operations
	#-------------------------------
	def __add__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.add(self, other) 
	
	def __radd__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.add(other, self)
	
	def __sub__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.subtract(self, other)
	
	def __rsub__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.subtract(other, self)
	
	def __mul__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.multiply(self, other) 
	
	def __rmul__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.multiply(other, self)
	
	def __truediv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.divide(self, other) 
	
	def __rtruediv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.divide(other, self)
	
	def __floordiv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.floor_divide(self, other) 
	
	def __rfloordiv__(self, other):
		if not self.has_same_axis_as(other):
			raise ValueError("Axes are not aligned for the operation.")
		return np.floor_divide(other, self)
	
	def __pow__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.power(self, other) 
	
	def __rpow__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.power(other, self)
	
	#===============================
	
	
	#-------------------------------
	# Basic comparison operations
	#-------------------------------
	def __eq__(self, other):
		return np.equal(self, other)
	
	def __ne__(self, other):
		return np.not_equal(self, other)
	
	def __lt__(self, other):
		return np.less(self, other)
	
	def __le__(self, other):
		return np.less_equal(self, other)
	
	def __gt__(self, other):
		return np.greater(self, other)
	
	def __ge__(self, other):
		return np.greater_equal(self, other)
	
	#===============================
	
	
	#-----------------------------------
	# Utility Methods
	#-----------------------------------
	
	def unpack(self):
		"""
		Unpacks the object into easy to work
		with data structures.

		Returns
		-------
		(np.ndarray, np.ndarray)
			- y: Signal data array.
			- x: Signal axis array.
		"""
		
		y = self.y.values
		x = self.x.values
		
		return (y, x)
	
	def copy(self) -> Self:
		"""
		Returns a new copy of the signal.

		Returns
		-------
		Self
			A new copy of the object.
		"""
		copied_y = self.y.copy()
		copied_x = self.x.copy()
		title = self.title # Immutable, hence no need to copy
		
		return self.__class__(y=copied_y, x=copied_x, title=title)
	
	def set_meta_info(self, title=None, y_label=None, x_label=None) -> None:
		"""
		Set meta info about the signal.

		Parameters
		----------
		title: str
			- Title for the signal
			- e.g. "Speedometer"
		y_label: str
			- Label for the y-axis.
			- e.g. "Speeed (m/s)"
		x_label: str
			- Label for the x-axis.
			- e.g. "Distance (m)"
		"""
		
		y, x = self.y, self.x
		
		new_title = str(title) if title is not None else self.title
		new_y_label = str(y_label) if y_label is not None else y.label
		new_x_label = str(x_label) if x_label is not None else x.label
		
		# We create a new copy of the data and axis
		new_y = y.__class__(values=y.values().copy(), label=y_label)
		new_x = x.__class__(values=x.values().copy(), label=x_label)
		
		return self.__class__(y=new_y, x=new_x, title=title)
	
	
	def is_same_as(self, other: Self) -> bool:
		"""
		Check if two `S1D` instances are equal.
		"""
		
		if not isinstance(other, type(self)):
			return False
		
		if not self.y.is_same_as(other.y):
			return False
		
		if not self.x.is_same_as(other.x):
			return False
		
		return True
	
	def has_same_axis_as(self, other) -> bool:
		"""
		Check if two 'S1D' instances have same
		axis. Many operations need to satify this.
		"""
		return self.x.is_same_as(other.x)
	
	def mask(self, condition, set_to=None) -> Self:
		"""
		Mask the signal based on condition and
		the values can be set.
		
		Parameters
		----------
		condition: Callable
			- Condition function to apply on values of the signal.
			- E.g. lambda x: x > 10
		set_to: Number
			- Number to replace the masked position values.

		Returns
		-------
		S1D
			Masked Signal
		"""
		
		mask = condition(self)
		new_val = set_to
		
		if set_to is None: # Return the mask as the same signal but with booleans
			return mask
		
		else:
			# We apply the mask and update the signal data
			new_data = self.y.mask(condition=condition, set_to=new_val)
			
			# Since we're just updating the data, there is no change in the axis
			return self.__class__(y=new_data, x=self.x.copy(), title=self.title)
	
	#===================================
	
	#-----------------------------------
	# Tools
	#-----------------------------------
	
	def plot(
		self,
		ax = None,
		fmt: str = "k-",
		title = None,
		y_label = None,
		x_label = None,
		y_lim = None,
		x_lim = None,
		highlight_regions = None,
		vlines = None,
		hlines = None,
		legend = None,
		show_grid = True,
		show_stem = False,
	) -> plt.Figure | None:
		"""
		Plot the signal.
		
		.. code-block:: python
		
			import modusa as ms
			import numpy as np
			signal = ms.S1D(data=np.random.random(100), data_label="My data (unit)", title="My Random Signal")
			display(signal.plot())
		
		Parameters
		----------
		ax : matplotlib.axes.Axes | None
			- If you want to plot the signal on a given matplotlib axes, you can pass the ax here. We do not return any figure in this case.
			- If not passed, we create a new figure, plots the signal on it and then return the figure.
		fmt : str
			- Format of the plot as per matplotlib standards (Eg. "k-" or "blue--o)
			- Default is "k-"
		title : str | None
			- Title for the plot.
			- If not passed, we use the default set during signal instantiation.
		y_label : str | None
			- Label for the y-axis.
			- If not passed, we use the default set during signal instantiation.
		x_label : str | None
			- Label for the x-axis.
			- If not passed, we use the default set during signal instantiation.
		y_lim : tuple[float, float] | None
			- Limits for the y-axis.
		x_lim : tuple[float, float] | None
			- Limits for the x-axis.
		highlight_regions : list[tuple[float, float, str]] | None
			- List of time intervals to highlight on the plot.
			- [(start, end, 'tag')]
		vlines: list[float]
			- List of x values to draw vertical lines.
			- e.g. [10, 13.5]
		hlines: list[float]
			- List of data values to draw horizontal lines.
			- e.g. [10, 13.5]
		show_grid: bool
			- If true, shows grid.
			- Default: True
		show_stem : bool
			- If True, use a stem plot instead of a continuous line. 
			- Autorejects if signal is too large.
		legend : str | tuple[str, str] | None
			- If provided, adds a legend at the specified location.
			- e.g., "signal" -> gets converted into ("signal", "best")
			- e.g. ("signal", "upper right")
		
		Returns
		-------
		matplotlib.figure.Figure | None
			- The figure object containing the plot.
			- None in case an axis is provided.

		See Also
		--------
		modusa.tools.plotter.Plotter
		"""
		
		from modusa.tools.plotter import Plotter
		
		y: Data = self._y
		x: SAx = self._x
		y_val = y._values
		x_val = x._values
		
		if y_label is None: y_label = self._y._label
		if x_label is None: x_label = self._x._label
		if title is None: title = self.title
		
		fig: plt.Figure | None = Plotter.plot_signal(y=y_val, x=x_val, ax=ax, fmt=fmt, title=title, y_label=y_label, x_label=x_label, y_lim=y_lim, x_lim=x_lim, highlight_regions=highlight_regions, vlines=vlines, hlines=hlines, show_grid=show_grid, show_stem=show_stem, legend=legend)
		
		return fig
	
	#===================================
	
	
	
	#----------------------------------
	# Information
	#----------------------------------
	
	def print_info(self) -> None:
		"""Prints info about the audio."""
		print("-" * 50)
		print(f"{'Title'}: {self._title}")
		print("-" * 50)
		print(f"{'Type':<20}: {self.__class__.__name__}")
		print(f"{'Shape':<20}: {self.shape}")
		
		# Inheritance chain
		cls_chain = " → ".join(cls.__name__ for cls in reversed(self.__class__.__mro__[:-1]))
		print(f"{'Inheritance':<20}: {cls_chain}")
		print("=" * 50)
	
	def __str__(self):
		y, x = self._y, self._x
		y_label = y._label
		y_val = y._values
		shape = self.shape
		
		
		arr_str = np.array2string(
			y_val,
			separator=", ",
			threshold=30,       # limit number of elements shown
			edgeitems=2,          # show first/last 3 rows and columns
			max_line_width=120,   # avoid wrapping
		)
		
		return f"{self._nickname}({arr_str})"
		
	def __repr__(self):
		y, x = self._y, self._x
		y_label = y._label
		y_val = y._values
		shape = self.shape
		
		arr_str = np.array2string(
			y_val,
			separator=", ",
			threshold=30,       # limit number of elements shown
			edgeitems=2,          # show first/last 3 rows and columns
			max_line_width=120,   # avoid wrapping
		)
		return f"{self._nickname}({arr_str})"
	#===================================


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .base import ModusaSignalData
from typing import Self, Any
import numpy as np

class Data(ModusaSignalData):
	"""
	Space to represent any modusa signal's data.

	Parameters
	----------
	values: array-like
		Data values as an array.
	label: str
		Label for the data.

	Note
	----
	- This is only for developers reference.
	- As a user of the library, you will never instantiate this class.
	- It is used internally and users will be provided APIs to interact with this class if necessary.	
	"""
	
	#--------Meta Information----------
	_name = "Data"
	_nickname = "data" # This is to be used in repr/str methods
	_description = "Space to represent any signal's data."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-27"
	#----------------------------------
	
	def __init__(self, values, label=None):
		super().__init__() # Instantiating `ModusaSignalData` class
		
		values = np.asarray(values)
		
		if values.ndim > 2:
			raise ValueError("We do not currently support, > 2 dim arrays, please create a GitHub issue")
		
		self._values = values
		self._label = label or ""
	
	#-----------------------------------
	# Properties
	#-----------------------------------
	
	@property
	def values(self) -> np.ndarray:
		return self._values
	
	@property
	def label(self) -> str:
		return self._label
	
	@property
	def shape(self) -> tuple:
		return self.values.shape
	
	@property
	def ndim(self) -> tuple:
		return self.values.ndim
	
	@property
	def size(self) -> int:
		return self.values.size
	
	#===================================

	#-------------------------------
	# Utility methods
	#-------------------------------
	
	def is_same_as(self, other) -> bool:
		"""
		Check if two data instances are same.

		Parameters
		----------
		other: Data
			Another Data object to compare with.
		
		Returns
		-------
		bool
			True if same ow False

		Note
		----
		- We check the shape and all the values.
		- We are not checking the labels for now.
		"""
		
		arr1 = np.asarray(self)
		arr2 = np.asarray(other)
		
		if isinstance(other, type(self)):
			return False
		if arr1.shape != arr2.shape:
			return False
		if not np.allclose(arr1, arr2):
			return False
		
		return True
	
	def copy(self) -> Self:
		"""
		Return a new copy of data object.
		
		Returns
		-------
		Data
			A new copy of the data object.
		"""
		copied_values = np.asarray(self).copy()
		copied_label = self.label
		
		return self.__class__(values=copied_values, label=copied_label)
	
	def set_meta_info(self, label):
		"""
		Set meta info for the data.
		
		Parameters
		----------
		label: str
			Label for the data (e.g. "Intensity (dB)").
		Returns
		-------
		Self
			A new Self instance with new label.
		"""
		
		if label is None:
			return self
		else:
			return self.__class__(values=self.values.copy(), label=label)
		
	def mask(self, condition, set_to=None) -> Self:
		"""
		Mask the data based on condition and
		the values can be set using the `set_to` argument.
		
		Parameters
		----------
		condition: Callable
			- Condition function to apply on values of the data.
			- E.g. lambda x: x > 10
		set_to: Number
			- Number to replace the masked position values.

		Returns
		-------
		Data
			Masked Data with either booleans as per the condition or with updated values.
		"""
		
		mask = condition(self)
		new_value = set_to
		
		if new_value is None: # Return the mask as the same signal but with booleans
			return mask
	
		else:
			# We apply the mask and update the signal data
			data_arr = self.values.copy() # We do not want to modify the data inplace
			data_arr[mask] = new_value
			
			updated_data = Data(values=data_arr, label=self.label)
			
			return updated_data
	
	def pad(self, left=None, right=None) -> Self:
		"""
		Pad the data from left or right.

		Parameters
		----------
		left: arraylike
			- What to pad to the left of the signal.
			- E.g. 1 or [1, 0, 1], np.array([1, 2, 3])
		right: arraylike
			- What to pad to the right of the signal.
			- E.g. 1 or [1, 0, 1], np.array([1, 2, 3])

		Returns
		-------
		TDS
			Padded signal.
		"""
		
		if right is None and left is None: # No padding applied
			return self
		
		values = np.asarray(self).copy()
		if self.ndim == 1: # 1D
			if isinstance(left, (int, float)): left = [left]
			if isinstance(right, (int, float)): right = [right]
			
			if left is not None:
				values = np.concatenate((left, values))
			if right is not None:
				values = np.concatenate((values, right))
				
			return self.__class__(values=values, label=self.label)
			
		elif self.ndim == 2: # 2D
			if isinstance(left, (int, float)): left = [left]
			if isinstance(right, (int, float)): right = [right]
			
			if left is not None:
				left = np.asarray(left)
				if left.ndim != 1:
					raise ValueError(f"left must be 1 dimension")
				left_cols = np.tile(left, (self.shape[0], 1))
				values = np.concatenate((left_cols, values), axis=1)
				
			if right is not None:
				right = np.asarray(right)
				if right.ndim != 1:
					raise ValueError(f"right must be 1 dimension")
				right_cols = np.tile(right, (self.shape[0], 1))
				values = np.concatenate((values, right_cols), axis=1)
			
			return self.__class__(values=values, label=self.label)
			
	
	#================================
	
	
	#-------------------------------
	# NumPy Protocol
	#-------------------------------
	def __array__(self, dtype=None) -> np.ndarray:
		return np.asarray(self.values, dtype=dtype)
	
	def __index__(self):
		return int(self.values)
	
	def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
		"""
		Provides operation support for the universal functions
		on the Data object.

		"""
		
		raw_inputs = [x.values if isinstance(x, type(self)) else x for x in inputs]
		
		# Call the actual ufunc
		result = getattr(ufunc, method)(*raw_inputs, **kwargs)
		
		if isinstance(result, (np.ndarray, np.generic)):
			return self.__class__(result, label=self.label)
		else:
			return result
	
	def __array_function__(self, func, types, args, kwargs):
		"""
		Additional numpy function support.
		"""
		from modusa.utils import np_func_cat as nfc
		
		if not all(issubclass(t, type(self)) for t in types):
			return NotImplemented
		
		# Not supporting concatenate like operations as axis any random axis can't be concatenated
		if func in nfc.CONCAT_FUNCS:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
		
		# Single signal input expected
		data = args[0]
		data_arr = np.asarray(data)
		if "keepdims" not in kwargs: # We keep keepdim to True by default so that we remain in the same signal space.
			kwargs["keepdims"] = True
		result_arr = func(data_arr, **kwargs)
		
		if func in nfc.REDUCTION_FUNCS:
			return self.__class__(values=result_arr, label=data.label)
		
		elif func in nfc.X_NEEDS_ADJUSTMENT_FUNCS:
			# You must define logic for adjusting x
			raise NotImplementedError(f"{func.__name__} requires x-axis adjustment logic.")
			
		else:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")

	
	#================================
	
	#-------------------------------
	# Indexing
	#-------------------------------
	
	def __getitem__(self, key):
		"""
		Return a sliced or indexed view of the data.
	
		Parameters
		----------
		key : int | slice | tuple
			Index to apply to the values.
	
		Returns
		-------
		Data
			A new Data object with sliced values and same metadata,
			with singleton dimensions restored to match axis alignment.
		"""
		if not isinstance(key, (int, slice, tuple)):
			raise TypeError(f"Invalid key type {type(key)}")
			
		if self.ndim > 2:
			raise ValueError("Support for upto 2 dim only")
			
		sliced_arr = self.values[key]
		
		# Case 1: 1D array
		if self.ndim == 1:
			if sliced_arr.ndim == 0: # We have 
				# We need to make it 1D
				sliced_arr = np.expand_dims(sliced_arr, axis=0)
			
		# Case 2: 2D array
		if self.ndim == 2:
			if sliced_arr.ndim == 0:
				# We need to make it 2D
				sliced_arr = np.expand_dims(sliced_arr, axis=0)
				sliced_arr = np.expand_dims(sliced_arr, axis=0)
			elif sliced_arr.ndim == 1:
				if isinstance(key, int):
					axis_of_interest = 0
				# Assumption is that key is tuple of len 2 with atleast one position being integer, that is our axis of interest
				elif isinstance(key, tuple) and len(key) == 2:
					if isinstance(key[0], int):
						axis_of_interest = 0
					elif isinstance(key[1], int):
						axis_of_interest = 1
				# We need to make it 2D but based on the axis
				sliced_arr = np.expand_dims(sliced_arr, axis=axis_of_interest)
		
		return self.__class__(values=sliced_arr, label=self.label)
	
	def __setitem__(self, key, value):
		"""
		Set values at the specified index.
	
		Parameters
		----------
		key : int | slice | array-like | boolean array | S1D
			Index to apply to the values.
		value : int | float | array-like
			Value(s) to set.
		"""
		if not isinstance(key, (int, slice, tuple)):
			raise TypeError(f"Invalid key type {type(key)}")
			
		self.values[key] = value  # In-place assignment
	
	#===============================
	
	
	#-------------------------------
	# Basic arithmetic operations
	#-------------------------------
	def __add__(self, other):
		return np.add(self, other) 
	
	def __radd__(self, other):
		return np.add(other, self)
	
	def __sub__(self, other):
		return np.subtract(self, other)
	
	def __rsub__(self, other):
		return np.subtract(other, self)
	
	def __mul__(self, other):
		return np.multiply(self, other) 
	
	def __rmul__(self, other):
		return np.multiply(other, self)
	
	def __truediv__(self, other):
		return np.divide(self, other) 
	
	def __rtruediv__(self, other):
		return np.divide(other, self)
	
	def __floordiv__(self, other):
		return np.floor_divide(self, other) 
	
	def __rfloordiv__(self, other):
		return np.floor_divide(other, self)
	
	def __pow__(self, other):
		return np.power(self, other) 
	
	def __rpow__(self, other):
		return np.power(other, self)
	
	
	#-------------------------------
	# Basic comparison operations
	#-------------------------------
	def __eq__(self, other):
		return np.equal(self, other)
	
	def __ne__(self, other):
		return np.not_equal(self, other)
	
	def __lt__(self, other):
		return np.less(self, other)
	
	def __le__(self, other):
		return np.less_equal(self, other)
	
	def __gt__(self, other):
		return np.greater(self, other)
	
	def __ge__(self, other):
		return np.greater_equal(self, other)
	
	#===============================
	
	#-------------------------------
	# Representation
	#-------------------------------
	
	def __str__(self):

		arr_str = np.array2string(
			np.asarray(self),
			separator=", ",
			threshold=30,       # limit number of elements shown
			edgeitems=3,          # show first/last 3 rows and columns
			max_line_width=120,   # avoid wrapping
		)
		
		return f"{self._nickname}({arr_str})"
	
	def __repr__(self):
		
		arr_str = np.array2string(
			np.asarray(self),
			separator=", ",
			threshold=30,       # limit number of elements shown
			edgeitems=3,          # show first/last 3 rows and columns
			max_line_width=120,   # avoid wrapping
		)
		
		return f"{self._nickname}({arr_str})"
	
	#===============================


#!/usr/bin/env python3

# This contains categorised NumPy functions
# This is useful while handling np operations on modusa signals

import numpy as np

SHAPE_PRESERVING_FUNCS = {
	np.sin, np.cos, np.tan, 
	np.sinh, np.cosh, np.tanh,
	np.arcsin, np.arccos, np.arctan,
	np.exp, np.expm1, np.log, np.log2, np.log10, np.log1p,
	np.abs, np.negative, np.positive,
	np.square, np.sqrt, np.cbrt,
	np.floor, np.ceil, np.round, np.rint, np.trunc,
	np.clip, np.sign,
	np.add, np.subtract, np.multiply, np.true_divide, np.floor_divide
}

REDUCTION_FUNCS = {
	np.sum, np.prod,
	np.mean, np.std, np.var,
	np.min, np.max,
	np.argmin, np.argmax,
	np.median, np.percentile,
	np.all, np.any,
	np.nanmean, np.nanstd, np.nanvar, np.nansum
}

CONCAT_FUNCS = {
	np.concatenate, np.stack, np.hstack, np.vstack, np.dstack,
	np.column_stack, np.row_stack
}

X_NEEDS_ADJUSTMENT_FUNCS = {
	np.diff,
	np.gradient,
	np.trim_zeros,
	np.unwrap,
	np.fft.fft, np.fft.ifft, np.fft.fftshift, np.fft.ifftshift,
	np.correlate, np.convolve
}





#!/usr/bin/env python3

from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .tds import TDS
from .data import Data
from .t_ax import TAx
from modusa.tools.math_ops import MathOps
from typing import Self, Any
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path

class Audio(TDS):
	"""
	Represents a 1D audio signal within modusa framework.

	Note
	----
	- Use :class:`~modusa.generators.audio.AudioGen` to instantiate this class.
	- Audio must be mono (1D numpy array).

	Parameters
	----------
	y: Data
		- Data object holding the main audio data.
	t: TAx
		- Time axis for the audio.
	title: str
		- Title for the signal.
		- Default: None => ''
		- e.g. "MySignal"
		- This is used as the title while plotting.
	"""

	#--------Meta Information----------
	_name = "Audio Signal"
	_nickname = "Audio"
	_description = ""
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-04"
	#----------------------------------
	
	@validate_args_type()
	def __init__(self, y, t, title = None):

		super().__init__(y=y, t=t, title=title) # Instantiating `TimeDomainSignal` class
	
	#----------------------------------
	# Utility Tools
	#----------------------------------
			
	def play(self, regions = None, title = None):
		"""
		Play the audio signal inside a Jupyter Notebook.
	
		.. code-block:: python
	
			import modusa as ms
			audio = ms.audio.from_youtube("https://www.youtube.com/watch?v=lIpw9-Y_N0g")
			audio.play(regions=[(10, 30, "1")])
	
		Parameters
		----------
		regions : list[tuple[float, float, str], ...] | None
			- [(start_time, end_time, 'tag'), ...] pairs in seconds specifying the regions to play.
			- e.g. [(1, 10, "Segement 1"), (20, 25, "Segment 2")]
			- Default to None for which the entire signal is played.
		title : str | None
			- Title for the player interface. 
			- Defaults to None, which means that the signal’s internal title is used.
	
		Returns
		-------
		IPython.display.Audio
			An interactive audio player widget for Jupyter environments.

		See Also
		--------
		:class:`~modusa.tools.audio_player.AudioPlayer`
		"""
		
		from modusa.tools.audio_player import AudioPlayer

		title = title if title is not None else self.title

		audio_player = AudioPlayer.play(y=self.y.values, sr=self.t.sr, t0=self.t.t0, regions=regions, title=title)
		
		return audio_player


#!/usr/bin/env python3

from modusa import excp
from functools import wraps
from typing import Any, Callable, Type
from inspect import signature, Parameter
from typing import get_origin, get_args, Union
import types


#----------------------------------------------------
# Safety check for plugin (apply method)
# Check if the input type, output type is allowed,
# Also logs plugin usage.
#----------------------------------------------------
def plugin_safety_check(
	validate_plugin_input: bool = True,
	validate_plugin_output: bool = True,
	track_plugin_usage: bool = True
):
	def decorator(func: Callable) -> Callable:
		@wraps(func)
		def wrapper(self, signal: Any, *args, **kwargs):
			
			if validate_plugin_input:
				if not hasattr(self, 'allowed_input_signal_types'):
					raise excp.AttributeNotFoundError(f"{self.__class__.__name__} must define `allowed_input_signal_types`.")
				
				if type(signal) not in self.allowed_input_signal_types:
					raise excp.PluginInputError(f"{self.__class__.__name__} must take input signal of type {self.allowed_input_signal_types} but got {type(signal)}")
			
			if track_plugin_usage:
				if not hasattr(signal, '_plugin_chain'):
					raise excp.AttributeNotFoundError(f"Signal of type {type(signal).__name__} must have a `_plugin_chain` attribute for plugin tracking.")
					
				if not isinstance(signal._plugin_chain, list):
					raise excp.TypeError(f"`_plugin_chain` must be a list, but got {type(signal._plugin_chain)}")
					
				signal._plugin_chain.append(self.__class__.__name__)
			
			result = func(self, signal, *args, **kwargs)
			
			if validate_plugin_output:
				if not hasattr(self, 'allowed_output_signal_types'):
					raise excp.AttributeNotFoundError(f"{self.__class__.__name__} must define `allowed_output_signal_types`.")
				if type(result) not in self.allowed_output_signal_types:
					raise excp.PluginInputError(f"{self.__class__.__name__} must return output of type {self.allowed_output_signal_types} but returned {type(result)}")
			return result
	
		return wrapper
	return decorator


#----------------------------------------------------
# Safety check for generators (generate method)
# Check if the ouput type is allowed.
#----------------------------------------------------
def generator_safety_check():
	"""
	We assume that the first argument is self, so that we can actually extract properties to
	validate.
	"""
	def decorator(func: Callable) -> Callable:
		@wraps(func)
		def wrapper(self, *args, **kwargs):
			result = func(self, *args, **kwargs)
			
			if not hasattr(self, 'allowed_output_signal_types'):
				raise excp.AttributeNotFoundError(
					f"{self.__class__.__name__} must define `allowed_output_signal_types`."
				)
			if type(result) not in self.allowed_output_signal_types:
				raise excp.PluginInputError(
					f"{self.__class__.__name__} must return output of type {self.allowed_output_signal_types}, "
					f"but returned {type(result)}"
				)
				
			return result
		return wrapper
	return decorator


#----------------------------------------------------
# Validation for args type
# When this decorator is added to a function, it
# automatically checks all the arguments with their
# expected types. (self, forward type references are
# ignored)
#----------------------------------------------------

def validate_arg(arg_name: str, value: Any, expected_type: Any) -> None:
	"""
	Checks if `value_type` matches `expected_type`.
	Raises TypeError if not.
	"""
	import types
	from typing import get_origin, get_args, Union
	
	origin = get_origin(expected_type)
	
	# Handle Union (e.g. int | None)
	if origin in (Union, types.UnionType):
		union_args = get_args(expected_type)
		for typ in union_args:
			typ_origin = get_origin(typ) or typ
			if isinstance(value, typ_origin):
				return
		
		# ❌ If none match
		expected_names = ", ".join(
			get_origin(t).__name__ if get_origin(t) else t.__name__ for t in union_args
		)
		raise excp.InputTypeError(
			f"Argument '{arg_name}' must be one of ({expected_names}), got {type(value).__name__}"
		)
	
	# Handle generic types like list[float], tuple[int, str]
	elif origin is not None:
		if not isinstance(value, origin):
			raise excp.InputTypeError(
				f"Argument '{arg_name}' must be of type {origin.__name__}, got {type(value).__name__}"
			)
		return
		
	# ✅ Handle plain types
	elif isinstance(expected_type, type):
		if not isinstance(value, expected_type):
			raise excp.InputTypeError(
				f"Argument '{arg_name}' must be of type {expected_type.__name__}, got {type(value).__name__}"
			)
		return
	# ❌ Unsupported type structure
	else:
		raise excp.InputTypeError(f"Unsupported annotation for '{arg_name}': {expected_type}")

def validate_args_type() -> Callable:
	def decorator(func: Callable) -> Callable:
		@wraps(func)
		def wrapper(*args, **kwargs):
			sig = signature(func)
			bound = sig.bind(*args, **kwargs)
			bound.apply_defaults()
			
			for arg_name, value in bound.arguments.items():
				param = sig.parameters[arg_name]
				expected_type = param.annotation
				
				# Skip unannotated or special args
				if expected_type is Parameter.empty or arg_name in ("self", "cls") or isinstance(expected_type, str):
					continue
				
				validate_arg(arg_name, value, expected_type)  # <- this is assumed to be defined elsewhere
				
			return func(*args, **kwargs)
		return wrapper
	return decorator

#-----------------------------------
# Making a property immutable
# and raising custom error message
# during attempt to modify the values
#-----------------------------------
def immutable_property(error_msg: str):
	"""
	Returns a read-only property. Raises an error with a custom message on mutation.
	"""
	def decorator(getter):
		name = getter.__name__
		private_name = f"_{name}"
		
		def setter(self, value):
			raise excp.ImmutableAttributeError(error_msg)
			
		return property(getter, setter)
	
	return decorator


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from modusa.generators.base import ModusaGenerator


class {class_name}(ModusaGenerator):
	"""

	"""
	
	#--------Meta Information----------
	_name = ""
	_description = ""
	_author_name = "{author_name}"
	_author_email = "{author_email}"
	_created_at = "{date_created}"
	#----------------------------------
	
	def __init__(self):
		super().__init__()
	


#!/usr/bin/env python3

from modusa import excp
from modusa.decorators import validate_args_type, immutable_property
from modusa.models.base import ModusaSignal
from abc import ABC, abstractmethod
from typing import Any

class ModusaGenerator(ABC):
	"""
	Base class for any type of signal generators for modusa framework.
	
	Note
	----
	- This class is intended to be subclassed by any Generator related tools built for the modusa framework.
	- In order to create a generator tool, you can use modusa-dev CLI to generate an generator template.
	- It is recommended to treat subclasses of ModusaGenerator as namespaces and define @staticmethods with control parameters, rather than using instance-level __init__ methods.
	
	
	"""
	
	#--------Meta Information----------
	_name = ""
	_description = ""
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-04"
	#----------------------------------
	


#!/usr/bin/env python3

import numpy as np
from IPython.display import display, HTML, Audio

def play(y: np.ndarray, sr: float, title = None, t0: float = 0.0, regions = None) -> None:
		"""
		Plays audio clips for given regions in Jupyter Notebooks.

		Parameters
		----------
		y : np.ndarray
			- Audio data.
			- Mono (1D) numpy array.
		sr: float
			- Sampling rate of the audio.
		title : str | None
			- Title to display above audio players.
		t0: float
			- Starting timestamp, incase the audio is cropped
			- Default: 0.0 → Starts from 0.0 sec
		regions : list[tuple[float, float, str]] | tuple[float, float, str] | None
			- Regions to extract and play (in sec), e.g. [(0, 10.2, "tag")]
			- If there is only one region, a tuple should also work. e.g. (0, 10.2, "tag")
			- Default: None → The entire song is selected.

		Returns
		-------
		None
		"""
		if title:
			display(HTML(f"<p style='font-size:15px; font-weight:600; text-decoration: underline; margin-left:10px'>{title}</p>"))
			
		clip_tags = []
		timings = []
		players = []
	
		if isinstance(regions, tuple): regions = [regions] # (10, 20, "Region 1") -> [(10, 20, "Region 1")]
	
		if regions is not None:
			for region in regions:
				assert len(region) == 3
				
				start_sec = region[0] - t0
				end_sec = region[1] - t0
				tag = region[2]
				
				start_sample, end_sample = int(start_sec * sr), int(end_sec * sr)
				clip = y[start_sample: end_sample]
				audio_player = Audio(data=clip, rate=sr)._repr_html_()
				
				clip_tags.append(f"<td style='text-align:center; border-right:1px solid #ccc; padding:6px;'>{tag}</td>")
				timings.append(f"<td style='text-align:center; border-right:1px solid #ccc; padding:6px;'>{start_sec:.2f}s → {end_sec:.2f}s</td>")
				players.append(f"<td style='padding:6px;'>{audio_player}</td>")
		else:
			audio_player = Audio(data=y, rate=sr)._repr_html_()
			
			start_sec = t0
			end_sec = t0 + y.shape[0] / sr
			
			clip_tags.append(f"<td style='text-align:center; border-right:1px solid #ccc; padding:6px;'>1</td>")
			timings.append(f"<td style='text-align:center; border-right:1px solid #ccc; padding:6px;'>{start_sec:.2f}s → {end_sec:.2f}s</td>")
			players.append(f"<td style='padding:6px;'>{audio_player}</td>")
			
		# Wrap rows in a table with border
		table_html = f"""
		<div style="display:inline-block; border:1px solid #ccc; border-radius:6px; overflow:hidden;">
			<table style="border-collapse:collapse;">
				<tr style="background-color:#f2f2f2;">
					<th style="text-align:left; padding:6px 12px;">Clip</th>
					{''.join(clip_tags)}
				</tr>
				<tr style="background-color:#fcfcfc;">
					<th style="text-align:left; padding:6px 12px;">Timing</th>
					{''.join(timings)}
				</tr>
				<tr>
					<th style="text-align:left; padding:6px 12px;">Player</th>
					{''.join(players)}
				</tr>
			</table>
		</div>
		"""
	
		return HTML(table_html)

		
		
		